from typing import Dict, Optional, Union, List, Tuple
from dataspree.inference.ai.models.runtimes.runtime import Runtime, DetectionRuntime, ClassificationRuntime

import numpy as np
import os
from logging import getLogger

from dataspree.inference.core.exceptions import InferenceError

try:
    import onnxruntime

    ONNX_RUNTIME_AVAILABLE = True
except Exception:  # noqa
    ONNX_RUNTIME_AVAILABLE = False

logger = getLogger(__name__)


class OnnxRuntime(Runtime):
    def __init__(self, runtime_parameters: Dict, model_file_path: str, model_parameters: Dict,
                 execution_provider: Optional[str]) -> None:

        # only log errors to console
        onnxruntime.set_default_logger_severity(3)

        run_options = onnxruntime.RunOptions()
        sess_options = onnxruntime.SessionOptions()

        # see https://github.com/microsoft/onnxruntime/issues/4829 for details
        sess_options.enable_mem_pattern = False

        sess_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL

        execution_providers: List[Union[str, Tuple[str, Dict]]] = (
            onnxruntime.get_available_providers()) if execution_provider is None else [execution_provider]

        # Compile provider options
        provider_options = None
        if 'openvino_device' in runtime_parameters.keys():
            provider_options = [{'device_type': runtime_parameters['openvino_device']}]

        if (provider_options is None and 'OpenVINOExecutionProvider' in onnxruntime.get_available_providers() and
                execution_providers.index('OpenVINOExecutionProvider') == 0):

            if any(['MYRIAD' in x for x in onnxruntime.capi._pybind_state.get_available_openvino_device_ids()]):  # noqa
                # onnxruntime.capi._pybind_state.set_openvino_device("MYRIAD_FP16") # deprecated !
                provider_options = [{'device_type': 'MYRIAD_FP16'}]
                # set GPU_FP16 execution if openvino is execution provider
            if any(['GPU' in x for x in onnxruntime.capi._pybind_state.get_available_openvino_device_ids()]):  # noqa
                provider_options = [{'device_type': 'GPU_FP16'}]

        # Fix for error:
        # (Non-zero status returned while running FusedConv node Conv_129), smart pallet project,
        # performance measurement container (and main container).
        for i_prov, provider in enumerate(execution_providers):
            if provider == 'CUDAExecutionProvider':
                execution_providers[i_prov] = ('CUDAExecutionProvider', {'cudnn_conv_algo_search': 'EXHAUSTIVE', })

        logger.info(f'Attempting to launch with execution providers {execution_providers}')
        try:
            sess = onnxruntime.InferenceSession(model_file_path, sess_options, providers=execution_providers,
                                                provider_options=provider_options)
            sess.disable_fallback()
        except onnxruntime.capi.onnxruntime_pybind11_state.InvalidProtobuf as e:  # noqa
            raise ValueError(f'Could not load InferenceSession: {e}')

        meta = sess._model_meta.custom_metadata_map  # noqa
        Runtime.__init__(self,
                         input_shape=tuple(sess.get_inputs()[0].shape),  # (batch x channels x h # x w)
                         model_parameters=model_parameters)

        self.sess = sess
        self.run_options = run_options
        self.sess_options = sess_options

        self.input_name = self.sess.get_inputs()[0].name
        self.output_names = [o.name for o in self.sess.get_outputs()]
        self.metadata = meta

    def predict(self, input_batch: np.ndarray) -> List[np.ndarray]:
        output_maps = self.sess.run(self.output_names, {self.input_name: input_batch}, run_options=self.run_options)
        return output_maps

    def destroy(self):
        pass


class OnnxRuntimeDetection(OnnxRuntime, DetectionRuntime):
    def __init__(self, runtime_parameters: Dict, model_file_path: str, model_parameters: Dict,
                 execution_provider: Optional[str]) -> None:
        OnnxRuntime.__init__(self, runtime_parameters, model_file_path, model_parameters, execution_provider)
        DetectionRuntime.__init__(self, rotation_regression='rotation' in self.metadata.get('output_mapping', ''),
                                  half_rotation='2*rotation' in self.metadata.get('output_mapping', ''))

    def predict(self, input_batch: np.ndarray) -> List[np.ndarray]:
        return OnnxRuntime.predict(self, input_batch)

    def destroy(self) -> None:
        OnnxRuntime.destroy(self)

    def compute_class_activation_map(self, prediction_output: List[np.ndarray]) -> np.ndarray:
        return prediction_output[0]


class OnnxRuntimeClassification(OnnxRuntime, ClassificationRuntime):

    def __init__(self, runtime_parameters, model_file_path, model_parameters, execution_provider,
                 show_heatmaps) -> None:

        if show_heatmaps:
            import onnx
            from onnx import helper, numpy_helper

            model = onnx.load(model_file_path)

            try:
                last_conv_name = None
                for i, n in enumerate([n for n in model.graph.node]):
                    if 'global' in n.name.lower() and 'pool' in n.name.lower():
                        last_conv_name = n.input[0]

                intermediate_layer_value_info = helper.ValueInfoProto()
                intermediate_layer_value_info.name = last_conv_name
                model.graph.output.extend([intermediate_layer_value_info])
                model_file_path = model_file_path.replace('.onnx', 'CAM.onnx')

                weights = model.graph.initializer

                model_metadata = {}
                [model_metadata.update({k.key: k.value}) for k in model.metadata_props]
                fc_layer_substring = model_metadata['classification_layer']
                weight_index = 0
                for i, n in enumerate(weights):
                    if fc_layer_substring + '.weight' in n.name:
                        weight_index = i
                self.last_fc_layer_weights = numpy_helper.to_array(weights[weight_index])
                onnx.save(model, model_file_path)
            except Exception:  # noqa
                show_heatmaps = False
                logger.warning('Error while preparing for CAM output. Disabling heatmaps')

        OnnxRuntime.__init__(self, runtime_parameters=runtime_parameters, model_file_path=model_file_path,
                             model_parameters=model_parameters, execution_provider=execution_provider)
        ClassificationRuntime.__init__(self)

        self.model_parameters = model_parameters
        self.show_heatmap: bool = show_heatmaps

        # clean up
        if self.show_heatmap:
            os.remove(model_file_path)

    def predict(self, input_batch: np.ndarray) -> List[np.ndarray]:
        return OnnxRuntime.predict(self, input_batch)

    def compute_class_activation_map(self, prediction_output: List[np.ndarray]):
        """
        Computes the Class Activation Map (CAM) from the given prediction output.

        Args:
            prediction_output (List[np.ndarray]): A list containing the following elements:
                - logits: np.ndarray of shape (n_roi, n_crop, n_tta, n_classes)
                - cam_convs: np.ndarray of shape (n_roi, n_crop, n_tta, d, o_h, o_w)

        Raises:
            InferenceError: If heatmaps are disabled (self.show_heatmap is False).

        Returns:
            np.ndarray: The computed CAM, normalized to the range [0, 1].

        """
        if not self.show_heatmap:
            raise InferenceError('Did not prepare for CAM output. Heatmaps are disabled.')

        logits, cam_conv = prediction_output[0], prediction_output[1]
        n_roi, n_crop, n_tta, n_classes = logits.shape
        d, o_h, o_w = cam_conv.shape[-3:]
        assert cam_conv.shape == (n_roi, n_crop, n_tta, d, o_h, o_w)

        cam_conv = cam_conv[:, :, :, np.newaxis]
        assert cam_conv.shape == (n_roi, n_crop, n_tta, 1, d, o_h, o_w)

        cam_weights = self.last_fc_layer_weights.view()
        attach_axes(cam_weights, 3, to=len(cam_conv.shape))
        assert cam_weights.shape[:5] == (1, 1, 1, n_classes, d) and len(cam_weights.shape) == len(cam_conv.shape)

        cams = (cam_weights * cam_conv).sum(axis=4)
        assert cams.shape == (n_roi, n_crop, n_tta, n_classes, o_h, o_w)
        cams = cams - cams.min()
        max_per_batch = np.take_along_axis(cams, attach_axes(np.argmax(logits, axis=3), to=len(cams.shape)), axis=3)
        max_per_batch.shape = (*max_per_batch.shape[:3], int(np.prod(max_per_batch.shape[3:])))
        return cams / attach_axes(max_per_batch.max(-1), 0, len(cams.shape))

    def destroy(self) -> None:
        OnnxRuntime.destroy(self)


def attach_axes(arr, before=0, to=0):
    arr.shape = (*((1,) * before), *arr.shape, *((1,) * (to - before - len(arr.shape))))
    return arr
