
from enum import Enum
from typing import Any, Dict, Tuple, List
from abc import ABC, abstractmethod

import numpy as np

OPENVINO_AVAILABLE: bool
try:
    # this avoids fallbacks to CPU if PATH is not correctly set, which happens when openvino is installed via pip
    # https://onnxruntime.ai/docs/execution-providers/OpenVINO-ExecutionProvider.html
    import openvino.utils as utils  # noqa

    utils.add_openvino_libs_to_path()
    OPENVINO_AVAILABLE = True
except Exception:  # noqa
    OPENVINO_AVAILABLE = False


class ExistingRuntimes(str, Enum):
    ONNX_RUNTIME = 'onnxruntime'
    TENSORRT = 'tensorrt'


class Runtime(ABC):

    def __init__(self, input_shape: Tuple[int, ...], model_parameters: Dict[Any, Any]) -> None:
        self.input_shape: Tuple[int, ...] = input_shape
        self.model_parameters: Dict[Any, Any] = model_parameters

    @abstractmethod
    def predict(self, input_batch: np.ndarray) -> List[np.ndarray]:
        ...

    @abstractmethod
    def destroy(self) -> None:
        ...


class DetectionRuntime(ABC):
    def __init__(self, rotation_regression: bool, half_rotation: bool) -> None:
        self.rotation_regression: bool = rotation_regression
        self.half_rotation: bool = half_rotation

        if self.half_rotation and not self.rotation_regression:
            raise ValueError('Half rotation implies rotation regression.')

    @abstractmethod
    def predict(self, input_batch: np.ndarray) -> List[np.ndarray]:
        ...

    @abstractmethod
    def destroy(self) -> None:
        ...

    @abstractmethod
    def compute_class_activation_map(self, prediction_output: List[np.ndarray]) -> np.ndarray:
        ...


class ClassificationRuntime(ABC):

    @abstractmethod
    def predict(self, input_batch: np.ndarray) -> List[np.ndarray]:
        ...

    @abstractmethod
    def destroy(self) -> None:
        ...

    @abstractmethod
    def compute_class_activation_map(self, prediction_output: List[np.ndarray]) -> np.ndarray:
        ...



