import time
from logging import getLogger

import os
from typing import List, Optional, Tuple
from collections import OrderedDict

import numpy as np

from dataspree.inference.ai.models.runtimes.runtime import Runtime, DetectionRuntime

TENSORRT_AVAILABLE: bool
try:
    import tensorrt as trt  # noqa
    import pycuda.driver as cuda  # noqa
    import pycuda.autoinit  # noqa

    # trt_logger = trt.Logger(trt.Logger.DEBUG)
    trt_logger = trt.Logger(trt.Logger.WARNING)

    TENSORRT_AVAILABLE = True
except Exception as e:  # noqa
    TENSORRT_AVAILABLE = False
    trt_logger = None

np.bool = np.bool_

logger = getLogger(__name__)


class TrtRuntime(Runtime):

    def __init__(self, runtime_parameters, model_file_path: Optional[str], model_parameters) -> None:
        if not TENSORRT_AVAILABLE:
            raise RuntimeError('Not available')

        cuda.init()
        device = cuda.Device(0)  # enter your Gpu id here
        self.cuda_context = device.make_context()

        self.runtime_parameters = runtime_parameters

        model_path_onnx: str = f'{os.path.splitext(model_file_path)[0]}.onnx'
        model_path_engine: str = f'{os.path.splitext(model_file_path)[0]}.engine'

        min_batch = self.runtime_parameters.get('min_batch', 1)
        opt_batch = self.runtime_parameters.get('opt_batch', 1)
        max_batch = self.runtime_parameters.get('max_batch', 1)

        image_input, output_shapes = self.build_engine(model_path_onnx, model_path_engine,
                                                       min_batch, opt_batch, max_batch)
        Runtime.__init__(self, input_shape=image_input, model_parameters=model_parameters)

        self.engine = self.load_engine(model_path_engine)
        self.context = self.engine.create_execution_context()
        self.output_shapes = output_shapes

        # Allocate buffers
        self.inputs, self.outputs, self.bindings, self.stream = self.allocate_buffers(self.engine, max_batch)

    def build_engine(self, model_path_onnx: str, model_path_engine: Optional[str],
                     min_batch: int, opt_batch: int, max_batch: int) -> (
            Tuple[Tuple[int, int, int, int], List[Tuple[int, ...]]]):
        builder = trt.Builder(trt_logger)
        config = builder.create_builder_config()

        cache = config.create_timing_cache(b'')
        config.set_timing_cache(cache, ignore_mismatch=False)
        # Enable dynamic batch shape
        flag = 1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
        network = builder.create_network(flag)
        parser = trt.OnnxParser(network, trt_logger)

        with open(model_path_onnx, 'rb') as f:
            if not parser.parse(f.read()):
                errors = []
                for error in range(parser.num_errors):
                    errors.append((parser.get_error(error)))
                raise OSError(f"ERROR: Failed to parse the ONNX file {model_path_onnx}: {errors}")

        # Enable dynamic batch shape
        network.get_input(0).shape = [-1, *network.get_input(0).shape[1:]]

        ins = [network.get_input(i) for i in range(network.num_inputs)]
        outs = [network.get_output(i) for i in range(network.num_outputs)]

        self.input_description = OrderedDict((k, v) for k, v in zip([i.name for i in ins],
                                                                    [i.shape for i in ins]))
        self.output_description = OrderedDict((k, v) for k, v in zip([i.name for i in outs],
                                                                     [i.shape for i in outs]))

        logger.info(f'Batch sizes: min={min_batch} opt={opt_batch} max={max_batch}.')

        if not os.path.isfile(model_path_engine):
            logger.info('Building Engine!')
            t0 = time.perf_counter()

            profile = builder.create_optimization_profile()

            # here, the input size is set to 1, but it should be set to the actual batch size (min, optimal, max)
            for name, shape in self.input_description.items():
                profile.set_shape(name, min=(min_batch, *shape[1:]),
                                  opt=(opt_batch, *shape[1:]),
                                  max=(max_batch, *shape[1:]))
            config.add_optimization_profile(profile)

            half = True
            int8 = False
            if half:
                config.set_flag(trt.BuilderFlag.FP16)
            elif int8:
                config.set_flag(trt.BuilderFlag.INT8)

            strip_weights = False
            if strip_weights:
                config.set_flag(trt.BuilderFlag.STRIP_PLAN)
            # config.set_flag(trt.kWEIGHT_STREAMING)

            engine_bytes = builder.build_serialized_network(network, config)

            with open(model_path_engine, "wb") as f:
                f.write(engine_bytes)
            logger.info(f'Building engine took {time.perf_counter() - t0} seconds.')

        return (-1, ins[0].shape[1], ins[0].shape[2], ins[0].shape[3]), [o.shape for o in outs]

    # @staticmethod
    def load_engine(self, engine_path: str):
        t0 = time.perf_counter()
        try:
            runtime = trt.Runtime(trt_logger)
            with open(engine_path, "rb") as f:
                return runtime.deserialize_cuda_engine(f.read())
        finally:
            logger.info(f'Loading the engine took {time.perf_counter() - t0} seconds.')

    @staticmethod
    def allocate_buffers(engine, max_batch_size: int):
        inputs, outputs, bindings = [], [], []
        stream = cuda.Stream()

        for i in range(engine.num_io_tensors):
            tensor_name = engine.get_tensor_name(i)

            size = trt.volume(engine.get_tensor_shape(tensor_name))
            dtype = trt.nptype(engine.get_tensor_dtype(tensor_name))

            # Allocate host and device buffers
            host_mem = cuda.pagelocked_empty(max_batch_size * abs(size), dtype)
            device_mem = cuda.mem_alloc(host_mem.nbytes)

            # Append the device buffer address to device bindings
            bindings.append(int(device_mem))

            # Append to the appropriate input/output list
            if engine.get_tensor_mode(tensor_name) == trt.TensorIOMode.INPUT:
                inputs.append((host_mem, device_mem))
            else:
                outputs.append((host_mem, device_mem))

        return inputs, outputs, bindings, stream

    def predict(self, input_batch: np.ndarray) -> List[np.ndarray]:
        host_memory, device_memory = self.inputs[0]

        # Transfer input data to device
        number_of_batches = input_batch.shape[0]

        raveled = input_batch.ravel()
        host_memory = host_memory[:len(raveled)]
        np.copyto(host_memory, raveled)

        for i, shape in self.input_description.items():
            new_shape = (number_of_batches, *shape[1:])
            self.context.set_input_shape(i, new_shape)

        cuda.memcpy_htod_async(device_memory, host_memory, self.stream)

        # Set tensor address
        for i in range(self.engine.num_io_tensors):
            self.context.set_tensor_address(self.engine.get_tensor_name(i), self.bindings[i])

        # Run inference.
        self.context.execute_async_v3(stream_handle=self.stream.handle)  # , batch_size = input_batch.shape[0])

        # Transfer predictions back from the GPU.
        for o in self.outputs:
            cuda.memcpy_dtoh_async(o[0], o[1], self.stream)

        for i in range(len(self.outputs)):
            cuda.memcpy_dtoh_async(host_memory, device_memory, self.stream)

        # Synchronize the stream
        self.stream.synchronize()

        return [output[0].reshape(self.output_shapes[oid])[:number_of_batches] for oid, output in enumerate(
            self.outputs)]

    def destroy(self) -> None:
        if self.cuda_context:
            self.cuda_context.pop()


class TrtRuntimeDetection(TrtRuntime, DetectionRuntime):

    def __init__(self, runtime_parameters, model_file_path, model_parameters) -> None:
        TrtRuntime.__init__(self, runtime_parameters, model_file_path, model_parameters)
        DetectionRuntime.__init__(self, rotation_regression='rotation' in model_parameters.get('output_mapping', ''),
                                  half_rotation='2*rotation' in model_parameters.get('output_mapping', ''))

    def predict(self, input_batch: np.ndarray) -> List[np.ndarray]:
        return TrtRuntime.predict(self, input_batch)

    def destroy(self) -> None:
        TrtRuntime.destroy(self)

    def compute_class_activation_map(self, prediction_output: List[np.ndarray]) -> np.ndarray:
        return prediction_output[0]
