import weakref
import numpy as np

from ctypes import Structure, c_char_p, c_int, c_bool, c_void_p, cdll, POINTER
from typing import Callable

from dataspree.inference.core.design_patterns import Singleton
import logging

from dataspree.inference.core.exceptions import DataspreeInferenceException

logger = logging.getLogger(__name__)


class NlabCtrlStepMotor(Structure):
    _fields_ = [
        ("id", c_char_p),
        ("name", c_char_p),
        ("step", c_int),
        ("min_step", c_int),
        ("max_step", c_int),
    ]


class NlabCtrlInfo(Structure):
    _fields_ = [
        ("backend_id", c_char_p),
        ("id", c_char_p),
        ("dev_path", c_char_p),
    ]


class NlabCtrlLed(Structure):
    _fields_ = [
        ("id", c_char_p),
        ("name", c_char_p),
        ("on", c_bool),
        ("strobe_on", c_bool),
        ("brightness", c_int),
        ("strobe_delay", c_int)
    ]


class NlabCtrlError(Structure):
    _fields_ = [
        ("code", c_int),  # is an enum
        ("msg", c_char_p)
    ]


class NlabCtrl(Structure):
    _fields_ = [
        ("go_ptr", c_void_p)
    ]


class NlabCtrlOpts(Structure):
    _fields_ = [
        ("state_dir", c_char_p)
    ]


class WahtariControllerException(Exception):
    pass


class WahtariController(Singleton):

    def initialize(self, path='libnlab-ctrl.so', controller_index=0, state_dir=b'/tmp/nlab-ctrl-state'):

        self.lib = cdll.LoadLibrary(path)
        self.c_ctrl_p = None
        self.c_info_list = None
        self.c_error = None
        self.c_error_p = None
        self.c_led_p = None
        self.c_zoom = None
        self.c_focus = None
        self.c_opts = NlabCtrlOpts(state_dir=c_char_p(state_dir))

        # "do init stuff"
        self.lib.nlab_ctrl_error_new.restype = POINTER(NlabCtrlError)
        self.lib.nlab_ctrl_list.restype = POINTER(POINTER(NlabCtrlInfo))
        self.lib.nlab_ctrl_info_list_at_index.restype = POINTER(NlabCtrlInfo)
        self.lib.nlab_ctrl_open.restype = POINTER(NlabCtrl)

        self.lib.nlab_ctrl_get_leds.restype = POINTER(POINTER(NlabCtrlLed))
        self.lib.nlab_ctrl_leds_at_index.restype = POINTER(NlabCtrlLed)
        self.lib.nlab_ctrl_leds_size.restype = c_int

        self.lib.nlab_ctrl_get_step_motors.restype = POINTER(POINTER(NlabCtrlStepMotor))
        self.lib.nlab_ctrl_step_motors_at_index.restype = POINTER(NlabCtrlStepMotor)

        self.c_error_p = self.lib.nlab_ctrl_error_new()
        self.c_info_list_p = self.lib.nlab_ctrl_list(self.c_error_p)
        self.check_error()

        self.c_info_list_size = self.lib.nlab_ctrl_info_list_size(self.c_info_list_p)
        if controller_index >= self.c_info_list_size:
            raise ValueError(
                f'Controller index {controller_index} is invalid. Number of controllers: {self.c_info_list_size}')

        self.c_info = self.lib.nlab_ctrl_info_list_at_index(self.c_info_list_p, controller_index)

        self.c_ctrl_p = self.lib.nlab_ctrl_open(self.c_info.contents.backend_id, self.c_info.contents.dev_path,
                                                self.c_opts,
                                                self.c_error_p)
        self.check_error()

        # leds
        self.c_leds = self.lib.nlab_ctrl_get_leds(self.c_ctrl_p, self.c_error_p)
        self.check_error()
        self.c_leds_size = self.lib.nlab_ctrl_leds_size(self.c_leds)

        # motors
        self.c_step_motors = None
        self.c_step_motors_size = 0
        self.update_zoom_and_focus()

        logger.info('Initialized Wahtari controller.')

    def update_zoom_and_focus(self):
        self.c_step_motors = self.lib.nlab_ctrl_get_step_motors(self.c_ctrl_p, self.c_error_p)
        self.check_error()
        self.c_step_motors_size = self.lib.nlab_ctrl_step_motors_size(self.c_step_motors)
        for i in range(self.c_step_motors_size):
            mtr = self.lib.nlab_ctrl_step_motors_at_index(self.c_step_motors, i)
            if mtr.contents.name == b'Zoom':
                self.c_zoom = mtr
            elif mtr.contents.name == b'Focus':
                self.c_focus = mtr
        self.check_error()

    def switch_led(self, index: int, on: bool = True):
        if index < self.c_leds_size:
            c_led_p = self.lib.nlab_ctrl_leds_at_index(self.c_leds, index)
            self.lib.nlab_ctrl_set_led(self.c_ctrl_p, c_led_p.contents.id, on, self.c_error_p)
            self.check_error()
        else:
            logger.warning(f'Index {index} out of range ({self.c_leds_size}).')

    def switch_all_leds(self, on: bool = True):
        for i in range(self.c_leds_size):
            self.switch_led(i, on)

    def set_led_brightness(self, index: int, brightness: int = 50):
        if index < self.c_leds_size:
            c_led_p = self.lib.nlab_ctrl_leds_at_index(self.c_leds, index)
            if 0 <= brightness <= 100:
                self.lib.nlab_ctrl_set_led_brightness(self.c_ctrl_p, c_led_p.contents.id, brightness, self.c_error_p)
                self.check_error()
            else:
                ValueError(f'Brightness {brightness} out of range [0, 100]. ')
        else:
            ValueError(f'Index {index} out of range ({self.c_leds_size}).')

    def get_led_brightness(self, index: int):
        self.c_leds = self.lib.nlab_ctrl_get_leds(self.c_ctrl_p, self.c_error_p)
        self.check_error()
        if index < self.c_leds_size:
            c_led_p = self.lib.nlab_ctrl_leds_at_index(self.c_leds, index)
            return c_led_p.contents.brightness
        else:
            ValueError(f'Index {index} out of range ({self.c_leds_size}).')

    def get_zoom_position(self):
        self.update_zoom_and_focus()
        if self.c_zoom is None:
            raise DataspreeInferenceException('c zoom is None')
        return self.c_zoom.contents.step

    def get_focus_position(self):

        self.update_zoom_and_focus()
        # TODO: Do you mean self.c_focus here? I did not change in order to not break anything.
        if self.c_zoom is None:
            raise DataspreeInferenceException('c zoom is None')
        return self.c_zoom.contents.step

    def set_zoom_step(self, step: int):
        if self.c_zoom is None:
            raise DataspreeInferenceException('c zoom is None')

        self.lib.nlab_ctrl_set_step_motor_rel_pos(self.c_ctrl_p, self.c_zoom.contents.id, step, self.c_error_p)

    def set_focus_step(self, step: int):
        if self.c_focus is None:
            raise DataspreeInferenceException('c focus is None')

        self.lib.nlab_ctrl_set_step_motor_rel_pos(self.c_ctrl_p, self.c_focus.contents.id, step, self.c_error_p)

    def get_number_of_leds(self):
        return self.c_leds_size

    def check_error(self):
        if self.c_error_p is None:
            raise WahtariControllerException('Error structure is not initialized')
        if self.c_error_p.contents.code != 0:
            raise WahtariControllerException(self.c_error_p.contents.msg)

    def release(self):
        # switch all LEDs to 'off'
        self.switch_all_leds(False)

        # free error
        if self.c_error_p is not None:
            self.lib.nlab_ctrl_error_free(self.c_error_p)

        # free ctrl info list
        if self.c_info_list_p is not None:
            self.lib.nlab_ctrl_info_list_free(self.c_info_list_p)

        # free leds
        if self.c_leds is not None:
            self.lib.nlab_ctrl_leds_free(self.c_leds)

        # free stepmotors
        if self.c_step_motors is not None:
            self.lib.nlab_ctrl_step_motors_free(self.c_step_motors)
        # self.check_error()

        # CLOSE
        if self.c_ctrl_p is not None:
            self.lib.nlab_ctrl_close(self.c_ctrl_p)

    def __call__(self, led_1_brightness: float, led_2_brightness: float, on: bool = True):

        if led_1_brightness > 0:
            self.set_led_brightness(index=0, brightness=int(np.clip(led_1_brightness, 0, 100)))
            self.switch_led(index=0, on=True)
        else:
            self.switch_led(index=0, on=False)

        if led_2_brightness > 0:
            self.set_led_brightness(index=1, brightness=int(np.clip(led_2_brightness, 0, 100)))
            self.switch_led(index=1, on=True)
        else:
            self.switch_led(index=1, on=False)
        return self

    def __enter__(self):
        return weakref.proxy(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.switch_all_leds(False)
