from typing import Optional, List, Dict, Any, Callable, Tuple, Union
from enum import Enum

import numpy as np


def get_rotation_matrix_2d(*, angle_in_deg: Optional[float] = None, angle_in_rad: Optional[float] = None) -> np.ndarray:
    """
    Return the 2 dimensional rotation matrix.
    :param angle_in_deg: Angle in degrees.
    :param angle_in_rad: Angle in radians.
    :return:  the two -dimensional rotation matrix.
    """
    if (angle_in_deg is None) == (angle_in_rad is None):
        raise ValueError(f'Pass angle either in degrees ({angle_in_deg}) in radians ({angle_in_rad}).')

    if angle_in_deg is not None:
        angle_in_rad = (angle_in_deg / 180.0) * np.pi

    return np.array([[np.cos(angle_in_rad), -1. * np.sin(angle_in_rad)],
                     [np.sin(angle_in_rad), np.cos(angle_in_rad)]])


def coordinate_transformation(spanning_points: np.ndarray) \
        -> Tuple[Callable[[Union[np.ndarray, List[np.ndarray]]], np.ndarray],
        Callable[[Union[np.ndarray, List[np.ndarray]]], np.ndarray]]:
    """
    Compute functions for conducting a coordinate transformation from the canonical euclidean coordinate system
    to the coordinate system spanned by the #spanning_points.

    :param spanning_points: Three points (A, B, C), each \\in \\R^2 that span a two-dimensional coordinate system
                            through the vectors BA, BC.

    :return:                Tuple: Function to conduct transformation into the new coordinate system,
                                   function to convert backwards.

                            Each function takes as input a numpy array of size 2 or (n x 2), and emits an array
                            of the input size.
    """
    number_of_spanning_points: int = spanning_points.shape[0]
    dimensionality: int = spanning_points.shape[1]
    assert spanning_points.shape == (number_of_spanning_points, dimensionality)
    assert dimensionality + 1 == number_of_spanning_points
    assert dimensionality == 2, 'The dimensionality of the points has to be two, otherwise slightly adapt and test.'

    new_origin: np.ndarray = spanning_points[1]
    assert new_origin.shape == (dimensionality,)

    base_system = (spanning_points[[0, 2]] - new_origin).T
    assert base_system.shape == (dimensionality, dimensionality)

    coordinate_transform_matrix = np.linalg.pinv(base_system)
    assert coordinate_transform_matrix.shape == (dimensionality, dimensionality)

    # TODO: Use einsum notation and omit .T's
    return (lambda a: coordinate_transform_matrix.dot((a - new_origin).T).T,
            lambda a: base_system.dot(a.T).T + new_origin)


def intersect(a1, a2, b1, b2) -> Optional[np.ndarray]:
    a2, b2 = a2 - a1, b2 - b1
    nominator, denominator = np.cross((b1 - a1), b2), np.cross(a2, b2)
    assert (denominator != 0)

    lambda_a, lambda_b = nominator / denominator, np.cross((b1 - a1), a2) / denominator
    badness: float = max(-i if i <= 0 else i - 1 if i >= 1 else 0 for i in (lambda_a, lambda_b))
    return b1 + max(min(lambda_b, 1.0), 0.0) * b2 if badness < 1e-11 else None


def clipper(box_corners: np.ndarray, image_size: np.ndarray, visualize: bool = False) \
        -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
    """
    Clip boxes at the edge of the image.

    :param box_corners:              The previous box corners that may exceed the region_of_interest.
    :param image_size:               An unrotated box that denotes the region of interest.
    :param visualize:                If this value is set and a debug build is executed (see README), the
                                     old boxes and clipped boxes are plotted in a debug pyplot window.
    :return:                         The clipped box corners.
    """
    image_size = np.array(image_size) - 1
    img_corners = np.array([[0, 0], [image_size[0], 0], image_size, [0, image_size[1]]])
    box_corners_tlbr = [(*(bc < 0), *(bc > image_size)) for bc in box_corners]

    # a) Add all image corners. Determine later if they are inside the box
    interesting: List[np.ndarray] = list(img_corners)

    # b) Determine the box corners that are inside the image's boundaries.
    interesting.extend((bo for i, bo in enumerate(box_corners) if not any(box_corners_tlbr[i])))

    # c) Determine intersections between box' corners and image's boundaries.
    for i in range(len(box_corners)):
        for ic in (j for j in range(4) if box_corners_tlbr[i][j] != box_corners_tlbr[i - 1][j]):
            pi = intersect(*box_corners[[i - 1, i]], *img_corners[[ic - 1, ic]])
            if pi is not None:
                interesting.append(pi)

    encode, decode = coordinate_transformation(box_corners[:3])
    encoded = [ei for i, ei in enumerate(encode(interesting)) if
               i >= len(img_corners) or (np.all(0 - 1e-12 <= ei) and np.all(ei <= 1 + 1e-12))]
    if not len(encoded):
        return None, None

    min_x, min_y = np.min(encoded, axis=0)
    max_x, max_y = np.max(encoded, axis=0)

    new_box_corners = decode(np.array([[min_x, min_y], [min_x, max_y], [max_x, max_y], [max_x, min_y]]))

    # TODO unreachable code according to cython
    # if __debug__ and visualize:
    #  import matplotlib.pyplot as plt
    #     plt.plot(*np.array((*new_box_corners, new_box_corners[0])).T, color='red')
    #     plt.plot(*np.array((*box_corners, box_corners[0])).T, color='black', linestyle='dashed')
    #     plt.show()

    return new_box_corners, np.array(interesting[len(img_corners):])


def get_angles_from_points(points: np.ndarray) -> np.ndarray:
    """
    Return the angles from a set of two-dimensional points

    :param points:      points (shape N x 2)
    :return:            angles of points (shape N)
    """
    number_of_points: int = points.shape[0]
    point_dimensionality: int = points.shape[1]
    assert points.shape == (number_of_points, point_dimensionality)
    assert point_dimensionality == 2, 'Can only get angles from two-dimensional points. '

    angles = np.arctan2(*points[:, ::-1].T)
    assert angles.shape == (number_of_points,)
    return angles


class BoxReconstruction(Enum):
    MIN = 0
    AVG = 1
    MAX = 2


def parallelogram_to_box(parallelogram_flat: np.ndarray, longer_axis_index: np.ndarray,
                         box_reconstruction_mode: BoxReconstruction) -> np.ndarray:
    """
    Convert a set of parallelograms into boxes.

    :param parallelogram_flat:  List of points that together form a set of zero-centered parallelograms
                                (shape: total_number_of_points x state_dimensionality (=2))
    :param longer_axis_index:   Indices of the points corresponding to the longer axes in each parallelogram.
    :return:                    Parallelogram points (number_of
    """

    number_of_points = parallelogram_flat.shape[0]
    dimensionality: int = parallelogram_flat.shape[1]
    assert parallelogram_flat.shape == (number_of_points, dimensionality)
    assert dimensionality == 2, 'This method of transforming parallelograms to boxes only supports two-dimensional ' \
                                'state-spaces.'
    assert number_of_points % 2 == 0, 'the flat parallelogram points have to contain two points per parallelogram.'

    number_of_parallelograms: int = longer_axis_index.shape[0]
    assert longer_axis_index.shape == (number_of_parallelograms,)
    assert 2 * number_of_parallelograms == number_of_points

    other_point_index: np.ndarray = ((longer_axis_index + 1) % 2) + np.arange(longer_axis_index.shape[0]) * 2
    assert other_point_index.shape == (number_of_parallelograms,)

    if box_reconstruction_mode == BoxReconstruction.MAX:
        # Change the points of the parallelogram to the "enclosing" rectangle.
        parallelogram_flat[longer_axis_index, 0] += np.abs(parallelogram_flat[other_point_index, 0])
    elif box_reconstruction_mode == BoxReconstruction.AVG:
        # Change the points of the parallelogram to rectanlge that is the average between "enclosing" and "enclosed".
        pass
    elif box_reconstruction_mode == BoxReconstruction.MIN:
        # Change the points of the parallelogram to the "enclosed" rectangle.
        parallelogram_flat[longer_axis_index, 0] -= np.abs(parallelogram_flat[other_point_index, 0])

    parallelogram_flat[other_point_index, 0] = 0

    assert (np.abs(parallelogram_flat[longer_axis_index, 1]) < 1e-5).all()
    return parallelogram_flat


def get_corner_points_from_box_tuples(box_tuples: np.ndarray) -> np.ndarray:
    """
    Compute box corners from a numpy array of top and right coordinates of boxes.
    :param box_tuples: Top and right point of each box. (Shape: number of boxes x {top, right} (2) x dimensionality (2)
    :return:           Box corners for each box (Shape: number of boxes x number of box corners (4) x 2).
    """
    number_of_boxes: int = box_tuples.shape[0]
    number_of_support_points: int = box_tuples.shape[1]
    state_dimension: int = box_tuples.shape[2]
    assert box_tuples.shape == (number_of_boxes, number_of_support_points, state_dimension)
    assert number_of_support_points == 2
    assert state_dimension == 2

    top, right = box_tuples[:, 0], box_tuples[:, 1]
    assert top.shape == (number_of_boxes, state_dimension)
    assert right.shape == (number_of_boxes, state_dimension)

    # @dev: The more elegant solution np.stack((k, i, k, i), axis=1) is 3 times slower:
    #
    # >>> timeit.timeit(lambda: np.array([k, i, k ,i]).swapaxes(0, 1), number=100000)
    # 0.3523232340012328
    # >>> timeit.timeit(lambda: np.stack((k, i, k, i), axis=1),        number=100000)
    # 0.9020683050002845
    #
    # The reshaping operation is necessary here in order to keep the correct dimensionality in case
    # there are no boxes.
    box_corners: np.ndarray = np.array([top - right, top + right, right - top, -right - top]).swapaxes(0, 1).reshape(
        (number_of_boxes, 4, state_dimension))
    return box_corners


def rot_mats_multi_rads(matrices: np.ndarray, rotation_rads: np.ndarray) -> np.ndarray:
    """
    Rotate covariance matrices by multiple rotations.

    :param matrices:       Matrices to be rotated (Shape: number_of_rotations, 2, 2)
                           Attention: The rotation is performed inplace; hence the content of the "points" is
                           overwritten.
                           If you would like to keep the unrotated matrices, call this function on a copy of the
                           matrices' numpy array.

    :param rotation_rads:  Angles along which points are rotated. (Shape: number_of_rotations)

    :return:               Rotated matrices (Shape: number_of_rotations, 2, 2) to offer
                           the user the ability to call this function with a "rvalue" or a copy.
    """
    number_of_rotations: int = matrices.shape[0]
    state_dimensionality: int = matrices.shape[1]
    assert matrices.shape == (number_of_rotations, state_dimensionality, state_dimensionality)
    assert rotation_rads.shape == (number_of_rotations,)

    cr: np.ndarray = np.cos(rotation_rads)
    sr: np.ndarray = np.sin(rotation_rads)
    rotation_matrix = np.array([[cr, -sr], [sr, cr]]).reshape((2, 2, number_of_rotations))
    assert rotation_matrix.shape == (state_dimensionality, state_dimensionality, number_of_rotations)

    # TODO unreachable code according to cython
    # if __debug__:
    #     __matrices = matrices.copy()

    np.einsum('tTR,RTS,sSR->Rts', rotation_matrix, matrices, rotation_matrix, out=matrices)

    # TODO unreachable code according to cython
    # if __debug__:
    #     for i in range(number_of_rotations):
    #         __alt = rotation_matrix[:, :, i].dot(__matrices[i]).dot(rotation_matrix[:, :, i].T)
    #         _difference = np.abs(__alt - matrices[i])
    #         assert np.all(_difference < 1e-9)

    assert matrices.shape == (number_of_rotations, state_dimensionality, state_dimensionality)
    return matrices


def rot_points_multi_rads(points: np.ndarray, rotation_rads: np.ndarray) -> np.ndarray:
    """
    Rotate the box points inplace (!) and return a reference to the first argument.

    :param points:         Points to be rotated (Shape: number_of_rotations, number_of_points_per_rotation, 2)
                           Attention: The rotation is performed inplace; hence the content of the "points" is
                           overwritten.
                           If you would like to keep the unrotated points, call this function on a copy of the points
                           numpy array.

    :param rotation_rads:  Angles along which points are rotated. (Shape: number_of_rotations)

    :return:               Rotated points (Shape: number_of_rotations, number_of_points_per_rotation, 2) to offer
                           the user the ability to call this function with a "rvalue" or a copy.
    """
    number_of_rotations: int = points.shape[0]
    number_of_points_to_rotate_per_rotation: int = points.shape[1]
    state_dimensionality: int = points.shape[2]
    assert points.shape == (number_of_rotations, number_of_points_to_rotate_per_rotation, state_dimensionality)
    assert rotation_rads.shape == (number_of_rotations,)

    cr: np.ndarray = np.cos(rotation_rads)
    sr: np.ndarray = np.sin(rotation_rads)
    rotation_matrix = np.array([[cr, -sr], [sr, cr]]).reshape((2, 2, number_of_rotations))
    assert rotation_matrix.shape == (state_dimensionality, state_dimensionality, number_of_rotations)

    # TODO unreachable code according to cython
    # if __debug__:
    #     __points = points.copy()

    np.einsum('rcR,Rxc->Rxr', rotation_matrix, points, out=points)
    assert points.shape == (number_of_rotations, number_of_points_to_rotate_per_rotation, state_dimensionality)

    # TODO unreachable code according to cython
    # if __debug__:
    #     for i in range(len(__points)):
    #         # noinspection PyUnreachableCode
    #         __alt = rotation_matrix[:, :, i].dot(__points[i].swapaxes(0, 1)).swapaxes(0, 1)
    #         _difference = np.abs(__alt - points[i])
    #         assert np.all(_difference < 1e-9)

    return points


def get_box_and_rotation_from_parallelogram_keypoints(
        parallelogram_keypoints: np.ndarray, *, box_reconstruction_mode: BoxReconstruction = BoxReconstruction.MAX,
        visualize: bool = False) -> Tuple[np.ndarray, np.ndarray]:
    """
    Transform parallelograms into a boxes, and return the boxes and their rotation.

    :param parallelogram_keypoints:     Keypoints of parallelograms (Shape number of parallelogram
                                                                          x {center, top, right} x 2).

    :param visualize:                   If this value is set and a debug build is executed (see README), the
                                        internals of this functions are visualized in a debug pyplot window.

    :return:                            The boxes (shape number of boxes x 4 x 2) and their rotation
                                        (shape: number of boxes).
    """
    parallelogram_keypoints: np.ndarray = parallelogram_keypoints.astype(np.float64, copy=False)
    number_of_parallelograms, number_of_keypoints, dimensionality = parallelogram_keypoints.shape
    assert dimensionality == 2, 'This function is only designed to work for two-dimensional boxes'
    assert number_of_keypoints == 3, 'For two dimensional boxes we need exactly three keypoints to represent a ' \
                                     'parallelogram'

    parallelogram_center: np.ndarray = np.expand_dims(parallelogram_keypoints[:, 0, :], 1)
    assert parallelogram_center.shape == (number_of_parallelograms, 1, dimensionality)

    # @dev: This copy is necessary because we need a flat point representation of this array, and otherwise we could
    #       inplace reshape the numpy array because of a stride mismatch.
    parallelogram_tuple = parallelogram_keypoints[:, 1:3].copy()
    assert parallelogram_tuple.shape == (number_of_parallelograms, 2, dimensionality)
    np.subtract(parallelogram_tuple, parallelogram_center, out=parallelogram_tuple)

    # TODO unreachable code according to cython
    # if __debug__:
    #     __parallelogram_keypoints_original = parallelogram_tuple.copy()

    # Create a flat loookup index for the longer axes
    longer_axis_flat = np.argmax(np.linalg.norm(parallelogram_tuple, axis=2), 1)
    assert longer_axis_flat.shape == (number_of_parallelograms,)
    longer_axis_flat = np.add(longer_axis_flat, np.arange(number_of_parallelograms) * 2, out=longer_axis_flat)
    assert longer_axis_flat.shape == (number_of_parallelograms,)

    # Get rotation angle
    parallelogram_tuple.shape = (number_of_parallelograms * 2, dimensionality)
    parallelogram_tuple_flat: np.ndarray = parallelogram_tuple.reshape((number_of_parallelograms * 2, dimensionality))
    assert parallelogram_tuple_flat.shape == (number_of_parallelograms * 2, dimensionality)

    angle_parallelogram_x = get_angles_from_points(parallelogram_tuple_flat[longer_axis_flat])
    assert angle_parallelogram_x.shape == (number_of_parallelograms,)

    # Rotate
    parallelogram_tuple.shape = (number_of_parallelograms, 2, dimensionality)
    __tmp = rot_points_multi_rads(points=parallelogram_tuple, rotation_rads=-angle_parallelogram_x)
    assert np.all(__tmp == parallelogram_tuple)

    # TODO unreachable code according to cython
    # if __debug__:
    #     __parallelogram_keypoints_axis_aligned = parallelogram_tuple.copy()

    # Transform
    parallelogram_tuple.shape = (number_of_parallelograms * 2, dimensionality)
    __tmp = parallelogram_to_box(parallelogram_tuple, longer_axis_flat, box_reconstruction_mode=box_reconstruction_mode)
    parallelogram_tuple.shape = (number_of_parallelograms, 2, dimensionality)
    assert np.all(__tmp == parallelogram_tuple)
    assert np.all(__tmp.flatten() == parallelogram_tuple.flatten())

    # TODO unreachable code according to cython
    # if __debug__:
    #     __parallelogram_keypoints_transformed_axis_aligned = parallelogram_tuple.copy()

    # Rotate back
    __tmp = rot_points_multi_rads(points=parallelogram_tuple, rotation_rads=angle_parallelogram_x)
    assert np.all(__tmp == parallelogram_tuple)
    assert parallelogram_tuple.shape == (number_of_parallelograms, 2, dimensionality)

    box_corners = get_corner_points_from_box_tuples(parallelogram_tuple)
    assert box_corners.shape == (number_of_parallelograms, 4, dimensionality)
    np.add(box_corners, parallelogram_center, out=box_corners)

    # We need to discern the new "right" position from this potentially augmented box in order to extract the correct
    # rotation angle.
    # As in our use-cases, the semantics of "top" are always more important than the semantics of "right".
    # Therefore, we compute the angle of "top" + pi / 2 to get the new angle of the "right" keypoint.
    # As a consequence, the new "right" point might have been the old "left" points due to rotation augmentations.
    # However, top stays top.
    rotations = get_angles_from_points(parallelogram_tuple[:, 0]) + np.pi / 2
    assert rotations.shape == (number_of_parallelograms,)

    # TODO unreachable code according to cython
    # if __debug__ and visualize:
    # import matplotlib.pyplot as plt
    #     for i in range(number_of_parallelograms):
    #         angle: float = angle_parallelogram_x[i] / np.pi * 180
    #         plt.figure(figsize=(10, 10))
    #
    #         plt.title(f'Parallelogram {i} with angle: {angle:.1f}')
    #
    #         #
    #         # Original parallelogram
    #
    #         # Center
    #         plt.scatter(*__parallelogram_keypoints_original[i, 0], color='orange', marker='x', s=10)
    #         plt.arrow(0, 0, *__parallelogram_keypoints_original[i, 0], color='orange')
    #
    #         # Right
    #         plt.scatter(*__parallelogram_keypoints_original[i, 1], color='black', marker='x', s=10)
    #         plt.arrow(0, 0, *__parallelogram_keypoints_original[i, 1], color='black')
    #
    #         #
    #         # Rotated
    #         #
    #         plt.scatter(*__parallelogram_keypoints_axis_aligned[i, 0], color='red', s=20)
    #         plt.scatter(*__parallelogram_keypoints_axis_aligned[i, 1], color='green', s=20)
    #
    #         #
    #         # Rotated adapted
    #         #
    #         plt.scatter(*__parallelogram_keypoints_transformed_axis_aligned[i, 0], color='red', s=30)
    #         plt.scatter(*__parallelogram_keypoints_transformed_axis_aligned[i, 1], color='green', s=30)
    #
    #         # original box
    #         top, right = __parallelogram_keypoints_original[i, :]
    #         p1 = top + right
    #         p2 = -top + right
    #         p3 = -top - right
    #         p4 = top - right
    #         poly_points = np.array([p1, p2, p3, p4, p1])
    #         plt.plot(*poly_points.T, linestyle='--', color='blue')
    #
    #         # Transformed box
    #         top, right = parallelogram_tuple[i, :]
    #         p1 = top + right
    #         p2 = -top + right
    #         p3 = -top - right
    #         p4 = top - right
    #         poly_points = np.array([p1, p2, p3, p4, p1])
    #         plt.plot(*poly_points.T, linestyle='--', color='red')
    #
    #         plt.xlim(-150, 150)
    #         plt.ylim(-150, 150)
    #         plt.grid()
    #         plt.show()
    #
    #         ...  # < This is a good instruction for a breakpoint

    return box_corners, rotations


def assemble_keypoints_after_shuffle(keypoints: List[np.ndarray], instances: List[np.ndarray],
                                     positions: List[np.ndarray], **kwargs: np.ndarray) -> Tuple:
    """
    Assemble keypoints after a flat representation of parallelogram keypoints have been shuffled (for instance
    because of augmentations that mutate the order of keypoints).

    :param keypoints:       flat keypoints array (shape: (number_of_parallelograms * 3) x 2).
    :param instances:       Association of single keypoints to parallelograms via unique indices
                            (shape: number_of_parallelograms, ).
    :param positions:       Positions of the different keypoints (center, top, right) in the flat keypoints array
                            (shape: number_of_parallelograms, ).
    :param kwargs:          Additional np.ndarrays associated to original keypoints that are to be shuffled according
                            to instances. Shape: (number_of_unique_instances x *-1)
    :return:                Tuple[sorted keypoints,
                                  new instance lookup, (shape: number_of_unique_instances, )
                                  transformed kwargs   (shape: number_of_unique_instances, *-1)]
    """
    number_of_position_measurements: int = len(instances)
    keypoint_dimensionality: int = 2

    instance_array: np.ndarray = np.array(instances)
    assert instance_array.shape == (number_of_position_measurements,)

    keypoints: np.ndarray = np.array(keypoints).reshape(number_of_position_measurements,
                                                        keypoint_dimensionality)  # [:, :keypoint_dimensionality]
    keypoint_dimensionality: int = keypoints.shape[1]
    assert keypoints.shape == (number_of_position_measurements, keypoint_dimensionality)

    positions: np.ndarray = np.array(positions, dtype='object')
    assert positions.shape == (number_of_position_measurements,)

    unique_instances = np.unique(instance_array)
    number_of_unique_instances: int = unique_instances.shape[0]
    assert number_of_unique_instances == number_of_position_measurements / 3

    sorted_keypoints: np.array = np.empty((number_of_unique_instances, 3, keypoint_dimensionality))
    sorted_kwarg_arrays: Dict[Any, np.ndarray] = {k: np.empty((number_of_unique_instances, *v.shape[1:]))
                                                  for k, v in kwargs.items()}
    to_original_index: np.array(number_of_unique_instances) = np.empty(number_of_unique_instances)
    assert all((v.shape[0] == number_of_unique_instances for v in kwargs.values()))

    for i, uiid in enumerate(unique_instances):
        position_map = {positions[index]: index for index in np.where(instance_array == str(uiid))[0]}
        sorted_keypoints[i] = [keypoints[position_map[key]] for key in ['center', 'top', 'right']]

        for key, values in kwargs.items():
            sorted_kwarg_arrays[key][i] = values[int(uiid)]

        to_original_index[i] = int(uiid)

    return (sorted_keypoints, to_original_index, *sorted_kwarg_arrays.values())


def dlds_to_keypoints(dlds_boxes: np.ndarray) -> np.ndarray:
    """
    :param dlds_boxes:      boxes in DLDS format (shape number of boxes x 5)
    :return:                keypoints of the boxes as a triple of numpy arrays in flat representation:
                            keypoints (shape: number_of_boxes *3, 2), instance lookup (shape: number_of_boxes * 3, ),
                            position lookup.(shape: number_of_boxes * 3, ).
    """
    number_of_boxes: int = dlds_boxes.shape[0]
    assert dlds_boxes.shape == (number_of_boxes, 5)

    centers, distances = dlds_boxes[:, :2], dlds_boxes[:, 2:4] / 2
    rotations = (dlds_boxes[:, 4] * (-1) + np.pi) % (2 * np.pi) - np.pi  # + np.pi
    assert centers.shape == (number_of_boxes, 2)
    assert distances.shape == (number_of_boxes, 2)
    assert rotations.shape == (number_of_boxes,)

    distance_difference: np.ndarray = np.array([distances * np.array([0, 1]), distances * np.array([1, 0])]) \
        .reshape((2, number_of_boxes, 2)).swapaxes(0, 1)
    assert distance_difference.shape == (number_of_boxes, 2, 2)
    __tmp = rot_points_multi_rads(points=distance_difference, rotation_rads=rotations)
    assert np.all(__tmp == distance_difference), 'Rotations are supposed to happen inplace, __tmp is therefore just ' \
                                                 'a view to distance_difference.'
    assert distance_difference.shape == (number_of_boxes, 2, 2)

    top = centers - distance_difference[:, 0]
    right = centers + distance_difference[:, 1]
    assert top.shape == (number_of_boxes, 2) and right.shape == (number_of_boxes, 2) and \
           centers.shape == (number_of_boxes, 2) and rotations.shape == (number_of_boxes,)

    keypoints = np.hstack((centers, top, right))
    keypoints.shape = (number_of_boxes * 3, 2)

    return keypoints


def keypoints_trivial_lookup(number_of_keypoints: int) -> Tuple[np.ndarray, np.ndarray]:
    """
    :param number_of_keypoints: The number of keypoints
    :return:                    Tuple of numpy arrays assigning keypoints trivially.
                                keypoints (shape: number_of_boxes * 3, 2),
                                instance lookup (shape: number_of_boxes * 3, ),
                                position lookup.(shape: number_of_boxes * 3, ).
    """
    assert number_of_keypoints % 3 == 0
    number_of_boxes: int = int(number_of_keypoints / 3)

    instance_lookup = np.repeat(np.arange(0, number_of_boxes), 3).astype(str)
    assert instance_lookup.shape == (number_of_boxes * 3,)

    position_lookup = np.repeat(np.array(['center', 'top', 'right'])[np.newaxis], number_of_boxes, axis=0).flatten()
    assert position_lookup.shape == (number_of_boxes * 3,)

    return instance_lookup, position_lookup


def dlds_to_box_corners(dlds_boxes: np.ndarray) -> np.ndarray:
    """
    Transform rotated boxes from the DLDS format into box corner format ({Top Left, Top Right, Bottom Left,
    Bottom Right} in arbitrary order) to box corners.

    :param dlds_boxes:          Shape: (number of boxes x 5)
    :return:                    Shape: (number_of_boxes x number of corners (=4) x dimensionality (=2))
    """
    number_of_boxes: int = dlds_boxes.shape[0]
    number_of_dlds_box_coordinates: int = dlds_boxes.shape[1]
    assert dlds_boxes.shape == (number_of_boxes, number_of_dlds_box_coordinates)
    assert number_of_dlds_box_coordinates == 5, 'Boxes in DLDS format must consist of exactly five points: ' \
                                                '(center_x, center_y, width, height, rotation).'

    # Construct unrotated boxes centered at zero.
    center, sizes, rotations = dlds_boxes[:, :2], dlds_boxes[:, 2:4] / 2, dlds_boxes[:, 4]
    assert center.shape == (number_of_boxes, 2)
    assert sizes.shape == (number_of_boxes, 2)
    assert rotations.shape == (number_of_boxes,)

    box: np.ndarray = np.array([sizes * np.array([-1, -1]), sizes * np.array([-1, 1]),
                                sizes * np.array([1, 1]), sizes * np.array([1, -1])]).reshape((4, number_of_boxes, 2)) \
        .swapaxes(0, 1)
    assert box.shape == (number_of_boxes, 4, 2)

    # Rotate boxes
    __tmp = rot_points_multi_rads(points=box, rotation_rads=(rotations * (-1)))
    assert np.all(__tmp == box)
    np.add(box, center[:, np.newaxis], out=box)

    assert box.shape == (number_of_boxes, 4, 2)
    return box


def box_corners_to_dlds(box_corners: np.ndarray, rotations: np.ndarray) -> np.ndarray:
    """
    Transform rotated box corners into the DLDS format.

    :param box_corners:     Rotated box corners (Shape number_of_boxes x 4 x dimensionality (2)).
    :param rotations:       Numpy array that contains the rotation by which the box corners are rotated
                            (Shape number_of_boxes)
    :return:                Boxes in DLDS format (Shape: number of boxes x {center_x, center_y,
                                                                            width, height, rotation}) = 5)
    """
    number_of_boxes: int = box_corners.shape[0]
    number_of_box_corners: int = box_corners.shape[1]
    state_dimensionality: int = box_corners.shape[2]
    assert box_corners.shape == (number_of_boxes, number_of_box_corners, state_dimensionality)
    assert rotations.shape == (number_of_boxes,)
    assert number_of_box_corners == 4
    assert state_dimensionality == 2

    centers = box_corners.mean(axis=1)[:, np.newaxis]
    assert centers.shape == (number_of_boxes, 1, state_dimensionality)

    rotated_box_corners = rot_points_multi_rads(points=box_corners - centers, rotation_rads=-rotations)
    assert rotated_box_corners.shape == (number_of_boxes, number_of_box_corners, state_dimensionality)

    widths, heights = np.abs(rotated_box_corners[:, 1, :] - rotated_box_corners[:, 3, :]).T

    boxes_dlds: np.ndarray = np.array([centers[:, 0, 0], centers[:, 0, 1], widths, heights, rotations / -1]) \
        .reshape((5, number_of_boxes)).swapaxes(1, 0)

    assert boxes_dlds.shape == (number_of_boxes, 5)
    return boxes_dlds


def keypoints_to_dlds(keypoints: np.ndarray) -> np.ndarray:
    """
    Transform keypoints to box corners. Box corners have arbitrary order.

    :param keypoints:       Keypoints (Shape: number of parallelograms x {center, top, right} x dimensionality (=2))
    :return:                Box corners (Shape: number_of_parallelograms x 4 x dimensionality (=2))
    """

    number_of_parallelograms, number_of_keypoints, dimensionality = keypoints.shape
    assert number_of_keypoints == 3
    assert dimensionality == 2

    dlds_boxes: np.ndarray = np.empty((number_of_parallelograms, 5), dtype=np.float32)

    # set center
    dlds_boxes[:, :2] = keypoints[:, 0]

    # Set vectors to points in keypoints (inplace)
    np.subtract(keypoints[:, 1:3], keypoints[:, :1], out=keypoints[:, 1:3])

    # set new size
    dlds_boxes[:, 3:1:-1] = (np.linalg.norm(keypoints[:, 1:3], axis=2) * 2)

    # We need to discern the new "right" position from this potentially augmented box in order to extract the correct
    # rotation angle.
    # As in our use-cases, the semantics of "top" are always more important than the semantics of "right".
    # Therefore, we compute the angle of "top" + pi / 2 to get the new angle of the "right" keypoint.
    # As a consequence, the new "right" point might have been the old "left" points due to rotation augmentations.
    # However, top stays top.
    dlds_boxes[:, 4] = -(get_angles_from_points(keypoints[:, 1]) + np.pi / 2)

    return dlds_boxes


def keypoints_to_box_corners(keypoints: np.ndarray) -> np.ndarray:
    """
    Transform keypoints to box corners. Box corners have arbitrary order.

    :param keypoints:       Keypoints (Shape: number of parallelograms x {center, top, right} x dimensionality (=2))
    :return:                Box corners (Shape: number_of_parallelograms x 4 x dimensionality (=2))
    """
    number_of_parallelograms, number_of_keypoints, dimensionality = keypoints.shape
    assert number_of_keypoints == 3
    assert dimensionality == 2

    parallelogram_center = keypoints[:, 0][:, True]
    assert parallelogram_center.shape == (number_of_parallelograms, 1, dimensionality)

    parallelogram_tuple = keypoints[:, 1:3] - parallelogram_center
    assert parallelogram_tuple.shape == (number_of_parallelograms, 2, dimensionality)

    box_corners = get_corner_points_from_box_tuples(parallelogram_tuple)

    assert box_corners.shape == (number_of_parallelograms, 4, dimensionality)
    return box_corners + parallelogram_center


def inference_boxes_to_corners(boxes: np.ndarray, size_factor: float = 1.0) \
        -> Tuple[np.ndarray, np.ndarray]:
    number_of_boxes: int = len(boxes)

    corners: np.ndarray = np.empty((number_of_boxes, 4, 2), dtype=np.float32)
    corners[:, :4] = boxes[:, :2][:, np.newaxis]
    assert corners.shape == (number_of_boxes, 4, 2)

    orientations: np.ndarray = (np.pi - boxes[:, 4]) % (2 * np.pi)
    assert orientations.shape == (number_of_boxes,)

    distances = np.zeros((number_of_boxes, 4, 2), dtype=np.float32)
    distances[:, 0, 0] = boxes[:, 2] / 2 * size_factor
    distances[:, 1, 1] = boxes[:, 3] / 2 * size_factor
    assert distances.shape == (number_of_boxes, 4, 2)

    # Rotate
    distances = rot_points_multi_rads(distances, orientations)

    # Determine rotated corners
    to_bottom = distances[:, 0]
    to_right = distances[:, 1]

    top_left = corners[:, 0]
    bottom_left = corners[:, 1]
    top_right = corners[:, 2]
    bottom_right = corners[:, 3]
    assert top_left.shape == (number_of_boxes, 2)
    assert top_right.shape == (number_of_boxes, 2)
    assert bottom_right.shape == (number_of_boxes, 2)
    assert bottom_left.shape == (number_of_boxes, 2)

    left = corners[:, :2]
    right = corners[:, 2:]
    top = corners[:, 0::2]
    bottom = corners[:, 1::2]
    assert top.shape == (number_of_boxes, 2, 2)
    assert bottom.shape == (number_of_boxes, 2, 2)
    assert right.shape == (number_of_boxes, 2, 2)
    assert left.shape == (number_of_boxes, 2, 2)

    np.subtract(top, to_bottom[:, np.newaxis], out=top)
    np.add(bottom, to_bottom[:, np.newaxis], out=bottom)
    np.subtract(left, to_right[:, np.newaxis], out=left)
    np.add(right, to_right[:, np.newaxis], out=right)
    return corners, distances


def surrounding_rectangle_from_corners(corners: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    number_of_boxes: int = len(corners)
    assert corners.shape == (number_of_boxes, 4, 2)

    min_values = np.min(corners, axis=1).astype(np.int32)
    max_values = np.max(corners, axis=1).astype(np.int32)
    assert min_values.shape == (number_of_boxes, 2)
    assert max_values.shape == (number_of_boxes, 2)

    min_x = min_values[:, 0]
    min_y = min_values[:, 1]
    max_x = max_values[:, 0]
    max_y = max_values[:, 1]
    assert min_x.shape == (number_of_boxes,)
    assert min_y.shape == (number_of_boxes,)
    assert max_x.shape == (number_of_boxes,)
    assert max_y.shape == (number_of_boxes,)

    return min_values, max_values


def intersections(corners_a: np.ndarray, distances_a: np.ndarray,
                  corners_b: np.ndarray, distances_b: np.ndarray) -> np.ndarray:
    """
    https://math.stackexchange.com/questions/1278665/how-to-check-if-two-rectangles-intersect-rectangles-can-be-rotated

    Box corners:


    :param corners_a: array containing the box corners (top left, bottom left, top right,
                      bottom right = A D B C)
                              A---B
                              |   |
                              |   |
                              D---C
    :param distances_a:
    :param corners_b: array containing the box corners (top left, bottom left, top right,
                      bottom right = E H F G)
                              E---F
                              |   |
                              |   |
                              H---G
    :param distances_b:
    :return:
    """
    number_of_boxes_a: int = corners_a.shape[0]
    number_of_boxes_b: int = corners_b.shape[0]
    assert corners_a.shape == (number_of_boxes_a, 4, 2)
    assert corners_b.shape == (number_of_boxes_b, 4, 2)

    adbc = corners_b[:, 0][np.newaxis, :, np.newaxis] - corners_a[:, np.newaxis]
    assert adbc.shape == (number_of_boxes_a, number_of_boxes_b, 4, 2)
    ehfg = corners_a[:, 0][:, np.newaxis, np.newaxis] - corners_b[np.newaxis]
    assert adbc.shape == (number_of_boxes_a, number_of_boxes_b, 4, 2)
    assert ehfg.shape == (number_of_boxes_a, number_of_boxes_b, 4, 2)

    assert distances_a.shape == (number_of_boxes_a, 4, 2)
    to_bottom_a = (-distances_a[:, 0] * 2) / np.linalg.norm(distances_a[:, 0] * 2, axis=1)[:, None] ** 2  # AD
    to_right_a = (-distances_a[:, 1] * 2) / np.linalg.norm(distances_a[:, 1] * 2, axis=1)[:, None] ** 2  # AB
    assert to_bottom_a.shape == (number_of_boxes_a, 2) and to_right_a.shape == to_bottom_a.shape

    assert corners_b.shape == (number_of_boxes_b, 4, 2)
    assert distances_b.shape == (number_of_boxes_b, 4, 2)
    to_bottom_b = (-distances_b[:, 0] * 2) / np.linalg.norm(distances_b[:, 0] * 2, axis=1)[:, None] ** 2  # EH
    to_right_b = (-distances_b[:, 1] * 2) / np.linalg.norm(distances_b[:, 1] * 2, axis=1)[:, None] ** 2  # EF
    assert to_bottom_b.shape == (number_of_boxes_b, 2) and to_right_b.shape == to_bottom_b.shape

    # Compute the projection of EFGH (boxes b) on AB; AD
    b_ab = np.einsum('abft,at->abf', ehfg, to_right_a)
    b_ad = np.einsum('abft,at->abf', ehfg, to_bottom_a)
    a_ef = np.einsum('abft,bt->abf', adbc, to_right_b)
    a_eh = np.einsum('abft,bt->abf', adbc, to_bottom_b)
    assert b_ab.shape == (number_of_boxes_a, number_of_boxes_b, 4)
    assert len({b_ab.shape, b_ad.shape, a_ef.shape, a_eh.shape}) == 1

    intersections = np.zeros((number_of_boxes_a, number_of_boxes_b), dtype=bool)
    for i_a in range(number_of_boxes_a):
        for i_b in range(number_of_boxes_b):
            val = b_ab[i_a, i_b]
            if (val < 0).all() or (val > 1).all():
                continue

            val = b_ad[i_a, i_b]
            if (val < 0).all() or (val > 1).all():
                continue

            val = a_ef[i_a, i_b]
            if (val < 0).all() or (val > 1).all():
                continue

            val = a_eh[i_a, i_b]
            if (val < 0).all() or (val > 1).all():
                continue

            intersections[i_a, i_b] = 1

    return intersections
