"""Module for handling data types and their designations in pixel format processing.

This module defines `DataTypes` and `CompoundDataTypes` classes, along with utilities
for managing data types and parsing their designations.
"""

from abc import ABCMeta, abstractmethod

from enum import Enum
from typing import Union, Optional, get_args
from logging import getLogger

from dataspree.inference.sensor.data.pixel_format.utils import all_implementations

logger = getLogger(__name__)


class DefaultDataType(Enum):
    """Enumeration for default data types."""

    UNSIGNED = ''  # resp. u
    SIGNED = 's'
    FLOAT = 'f'


_custom_data_types: list[str] = []

DataType = Union[DefaultDataType, str]


def is_data_type(potential_data_type):
    """Checks if the given object is a valid data type.

    Args:
        potential_data_type (Any): The object to check.

    Returns:
        bool: True if the object is a valid data type, False otherwise.
    """
    for arg in get_args(DataType):
        if isinstance(potential_data_type, arg):
            return True

    return False


_reverse_data_type_mapping: dict[str, DataType] = {f.value: f for f in DefaultDataType}
_reverse_data_type_mapping['u'] = DefaultDataType.UNSIGNED


def datatype_name(data_type: DataType, multiple_designators: bool = False) -> str:
    """Returns the string representation of a data type.

    Args:
        data_type (DataType): The data type.
        multiple_designators (bool): Whether to include the 'u' designator for unsigned.

    Returns:
        str: The string representation of the data type.
    """
    if isinstance(data_type, DefaultDataType):
        if multiple_designators and data_type is DefaultDataType.UNSIGNED:
            return 'u'
        return data_type.value
    return data_type


def register_custom_data_type(name: str) -> None:
    """Registers a custom data type.

    Args:
        name (str): The custom data type name.

    Raises:
        ValueError: If the name is already registered or invalid.
    """
    if name in _custom_data_types:
        raise ValueError
    if name in _reverse_data_type_mapping:
        raise ValueError
    if len(name) != 1:
        raise ValueError
    _custom_data_types.append(name)


def data_types() -> list[DataType]:
    """Retrieves the list of available data types.

    Returns:
        List[DataType]: A list of all registered data types.
    """
    combined_list: list[DataType] = list(_custom_data_types)
    combined_list.extend(list(DefaultDataType))
    return combined_list


class DataTypes(metaclass=ABCMeta):
    """Abstract base class for data type designations."""

    @abstractmethod
    def designation(self) -> str:
        """Returns the string representation of the data type."""

    @abstractmethod
    def __len__(self) -> int:
        """Returns the number of elements in the data type."""

    @abstractmethod
    def __getitem__(self, item):
        """Returns the element at the specified index."""

    @staticmethod
    def create(full_designation: str, number_of_channels: int) -> list[tuple['DataTypes', str]]:
        """Creates instances of `DataTypes` based on a designation string.

        Args:
            full_designation (str): The full designation string.
            number_of_channels (int): The number of channels.

        Returns:
            List[Tuple[DataTypes, str]]: A list of parsed `DataTypes` instances and remaining designations.
        """
        compos: list[tuple[DataTypes, str]] = []
        for implementation in all_implementations(DataTypes):
            try:
                compos.extend(implementation.from_designation(full_designation, number_of_channels))
            except Exception as e:
                logger.warning(f'An exception occurred: {e} {implementation}')
        return compos

    @staticmethod
    @abstractmethod
    def from_designation(full_designation: str, number_of_channels: int) -> list[tuple['DataTypes', str]]:
        """Parses a designation string into `DataTypes` instances.

        Args:
            full_designation (str): The full designation string.
            number_of_channels (int): The number of channels.

        Returns:
            List[Tuple[DataTypes, str]]: A list of parsed `DataTypes` instances and remaining designations.
        """


class CompoundDataTypes(DataTypes):
    """Represents compound data types consisting of multiple elements."""

    def __init__(self, data: Optional[Union[list[DataType], DataType]]) -> None:
        """Initializes a `CompoundDataTypes` instance.

        Args:
            data (Optional[Union[List[DataType], DataType]]): The data types, or a single data type.
        """
        if data is None:
            data = DefaultDataType.UNSIGNED

        self.data: list[DataType]

        if isinstance(data, list):
            self.data = data
        else:
            assert is_data_type(data)
            self.data = [data]

    def designation(self) -> str:
        """Returns the string representation of the compound data type.

        Returns:
            str: The compound data type designation.
        """
        if not len(self.data):
            return ''

        if len(set(self.data)) == 1:
            return datatype_name(self.data[0])

        return ''.join(datatype_name(s) for s in self.data)

    def __len__(self) -> int:
        """Returns the number of elements in the compound data type.

        Returns:
            int: The number of elements.
        """
        return len(self.data)

    def __getitem__(self, item):
        """Returns the data type at the specified index.

        Args:
            item (int): The index.

        Returns:
            DataType: The data type at the index.
        """
        return self.data[item]

    def __eq__(self, other):
        """Checks equality with another object.

        Args:
            other (object): The object to compare.

        Returns:
            bool: True if equal, False otherwise.
        """
        return isinstance(other, CompoundDataTypes) and self.data == other.data

    @staticmethod
    def from_designation(full_designation: str, number_of_channels: int) -> list[tuple[DataTypes, str]]:
        """Parses a designation string into compound data types.

        Args:
            full_designation (str): The full designation string.
            number_of_channels (int): The number of channels.

        Returns:
            List[Tuple[DataTypes, str]]: A list of parsed `CompoundDataTypes` and remaining designations.
        """
        # Compound data type names don't contain numbers
        possible_data_types: list[tuple[DataTypes, str]] = [
            (CompoundDataTypes(DefaultDataType.UNSIGNED), full_designation)
        ]

        # Is the first letter a data type?

        decoded_data_types: list[DataType] = []
        for n in range(number_of_channels):
            if n >= len(full_designation):
                break

            data_type_designation: str = full_designation[n]
            if data_type_designation in _reverse_data_type_mapping:
                decoded_data_types.append(_reverse_data_type_mapping[data_type_designation])
            else:
                break

        if len(decoded_data_types) == number_of_channels:
            possible_data_types.append((CompoundDataTypes(decoded_data_types), full_designation[number_of_channels:]))

        if len(decoded_data_types) > 1:
            possible_data_types.append((CompoundDataTypes(decoded_data_types[0]), full_designation[1:]))

        return possible_data_types
