"""Contains sensor callbacks.

This module provides abstract base classes and utilities for managing sensor callbacks and controlling copy
behavior of class instances.
"""

from abc import ABCMeta, abstractmethod
from typing import TypeVar, Generic

from dataspree.inference.sensor.data import Sensor

U = TypeVar('U')


def copy_ignored(cls: type[U]) -> type[U]:
    """Decorator to make class instances only return `None` when copied or deep-copied.

    This effectively disables copying and deep-copying of instances of the decorated class.

    Args:
        cls (Type[T]): The class to be decorated.

    Returns:
        Type[T]: The decorated class with disabled copy functionality.
    """
    cls.__copy__ = lambda self: None  # type: ignore
    cls.__deepcopy__ = lambda self, memo: None  # type: ignore
    return cls


T = TypeVar('T', bound=Sensor)


@copy_ignored
class SensorCallback(Generic[T], metaclass=ABCMeta):
    """Abstract base class for handling various sensor events."""

    @abstractmethod
    def on_sensor_tracked(self, sensor_identifier: str) -> None:
        """Is evoked when a sensor is tracked.

        Args:
            sensor_identifier (str): The identifier of the tracked sensor.

        Raises:
            DriverFeatureImplementationError: If a listener encounters an error during execution.
        """

    @abstractmethod
    def on_sensor_found(self, sensor: T) -> None:
        """Is evoked when a sensor is found.

        Args:
            sensor (T): The sensor that has been found.

        Raises:
            DriverFeatureImplementationError: If a listener encounters an error during execution.
        """

    @abstractmethod
    def on_sensor_lost(self, sensor: T) -> None:
        """Is evoked when a sensor is lost.

        Args:
            sensor (T): The sensor that has been lost.

        Raises:
            DriverFeatureImplementationError: If a listener encounters an error during execution.
        """

    @abstractmethod
    def on_sensor_connected(self, sensor: T) -> None:
        """Is evoked when a sensor is connected.

        Args:
            sensor (T): The sensor that has been connected.

        Raises:
            DriverFeatureImplementationError: If a listener encounters an error during execution.
        """

    @abstractmethod
    def on_sensor_disconnected(self, sensor: T) -> None:
        """Is evoked when a sensor is disconnected.

        Args:
            sensor (T): The sensor that has been disconnected.

        Raises:
            DriverFeatureImplementationError: If a listener encounters an error during execution.
        """

    @abstractmethod
    def on_sensor_started(self, sensor: T) -> None:
        """Is evoked when a sensor has started.

        Args:
            sensor (T): The sensor that has started.

        Raises:
            DriverFeatureImplementationError: If a listener encounters an error during execution.
        """

    @abstractmethod
    def on_sensor_stopped(self, sensor: T) -> None:
        """Is evoked when a sensor has stopped.

        Args:
            sensor (T): The sensor that has stopped.

        Raises:
            DriverFeatureImplementationError: If a listener encounters an error during execution.
        """

    @abstractmethod
    def on_sensor_updated(self, sensor: T) -> None:
        """Is evoked when a sensor is updated.

        Args:
            sensor (T): The sensor that has been updated.

        Raises:
            DriverFeatureImplementationError: If a listener encounters an error during execution.
        """
