"""This module defines data structures for representing sensor models and descriptors."""

from typing import Optional
from dataclasses import dataclass

from dataspree.inference.sensor.data.configuration import NotAvailable, is_available


@dataclass
class SensorModel:
    """Contains all information relevant to generic sensors."""


@dataclass(frozen=True)
class SensorModelDescriptor:
    """Describes a sensor model with vendor, model, and optional identifier.

    Attributes:
        camera_vendor (Optional[str]): The vendor name of the camera.
        camera_model (Optional[str]): The model name of the camera.
        camera_id (Optional[str]): The unique identifier for the camera, if applicable.
    """

    camera_vendor: Optional[str]
    camera_model: Optional[str] = None
    camera_id: Optional[str] = None

    def __post_init__(self) -> None:
        """Validates the descriptor attributes for consistency."""
        if self.camera_id is not None:
            if self.camera_model is None or self.camera_vendor is None:
                raise ValueError('You need to specify vendor and model if you specify the ' 'camera identifier.')

        if self.camera_model is not None:
            if self.camera_vendor is None:
                raise ValueError('You need to specify vendor if you specify the camera identifier.')

    @property
    def description_level(self) -> int:
        """Calculates the description level based on provided attributes.

        Returns:
            int: The description level (0 to 3) indicating the specificity of the descriptor.
        """
        if self.camera_id is not None:
            return 3

        if self.camera_model is not None:
            return 2

        if self.camera_vendor is not None:
            return 1

        return 0

    def score(
        self, camera_id: str | NotAvailable, camera_model: str | NotAvailable, camera_vendor: str | NotAvailable
    ) -> int:
        """Calculates a matching score for the descriptor against provided camera details.

        Args:
            camera_id (Optional[str]): The identifier of the camera.
            camera_model (Optional[str]): The model of the camera.
            camera_vendor (Optional[str]): The vendor of the camera.

        Returns:
            int: A score indicating the specificity match level, with higher scores indicating a better match.
        """
        description_level = self.description_level
        generality_score: int = -description_level

        if self.camera_id is not None and self.camera_id == camera_id:
            return 300 + generality_score

        if (
            self.camera_model is not None
            and is_available(camera_model)
            and camera_model.lower().startswith(self.camera_model.lower())
        ):
            return 200 + generality_score

        if (
            self.camera_vendor is not None
            and is_available(camera_vendor)
            and self.camera_vendor.lower() == camera_vendor.lower()
        ):
            return 100 + generality_score

        return generality_score

    def __str__(self) -> str:
        """Generates a string representation of the sensor model descriptor.

        Returns:
            str: A human-readable description of the camera descriptor.
        """
        ident = '' if self.camera_id is None else f'{self.camera_id} '
        model = '' if self.camera_model is None else f'model "{self.camera_model}" '
        vendor = '' if self.camera_vendor is None else f'from vendor "{self.camera_vendor}"'
        descriptor = f'{ident}{model}{vendor}'
        return f'Camera Descriptor {descriptor}' if descriptor else 'Default Camera Descriptor'
