import traceback
from abc import abstractmethod, ABCMeta
from typing import Generic, Optional, TypeVar, Union, Callable, Any
from collections.abc import ItemsView
from collections.abc import Iterator, ValuesView, KeysView
from logging import getLogger

from dataspree.inference.core.environment import is_application_debug
from dataspree.inference.core.item import Item
from dataspree.inference.sensor.data import SensorState
from dataspree.inference.sensor.data.sensor import Sensor, SensorData, NotAvailable, TemperatureControl
from dataspree.inference.sensor.data.sensor_callback import SensorCallback

from dataspree.inference.sensor.data.exceptions import (
    DriverFeatureNotImplemented,
    DriverInvalidState,
    SensorNotFound,
    InputException,
    DriverFeatureImplementationError,
    SensorInvalidState,
    ConfigurationError,
    InvalidState,
    ResourceError,
    SensorFeatureUnavailable,
    SensorFeatureTemporarilyNotAvailable,
)

T = TypeVar('T', bound=Sensor)

logger = getLogger(__name__)


class Sensors(Generic[T]):
    """A class representing a collection of sensors, with methods to manage them.

    Attributes:
        __cb (SensorCallback): The callback function for sensor events.
        __values (Dict[str, Sensor]): A dictionary storing sensors by their identifier.
    """

    def __init__(self, cb: 'SensorCallback') -> None:
        """Initializes the Sensors class with a callback function.

        Args:
            cb (SensorCallback): The callback function for handling sensor events.
        """
        self.__cb: SensorCallback = cb
        self.__values: dict[str, T] = {}

    def add_sensor(self, sensor: T) -> None:
        """Adds a sensor to the collection and registers its callback.

        Args:
            sensor (T): The sensor to add.

        Raises:
            ValueError: If the sensor is already added or if its state is not DISCOVERED.

        Raises:
            InputException: If an exception is raised from the callbacks.

            InvalidState (InputException): If a callback is already registered.
        """
        if sensor.identifier in self.__values:
            raise DriverInvalidState(f'Could not add sensor: Sensor {sensor} already added.')

        if sensor.state != SensorState.DISCOVERED:
            raise SensorInvalidState(f'Could not add sensor: Sensor {sensor} is in an invalid state')

        sensor.register_callback(self.__cb)
        self.__values[sensor.identifier] = sensor

        # Evoke callbacks.
        callback_name: str = 'sensor_tracked'
        try:
            self.__cb.on_sensor_tracked(sensor.identifier)
            callback_name = 'sensor_found'
            self.__cb.on_sensor_found(sensor)
        except InputException as e:
            raise e from None
        except Exception as e:
            raise DriverFeatureImplementationError(
                f'Could not add sensor: Callback {callback_name} raised invalid exception: {type(e)} {e}'
            ) from e

    def __iter__(self) -> Iterator[str]:
        """Returns an iterator over sensor identifiers.

        Returns:
            Iterator[str]: An iterator over the keys of the sensors dictionary.
        """
        return self.__values.__iter__()

    def __len__(self) -> int:
        """Returns the number of sensors.

        Returns:
            int: The number of sensors in the collection.
        """
        return self.__values.__len__()

    def __getitem__(self, item: str) -> T:
        """Gets a sensor by identifier.

        Args:
            item (str): The identifier of the sensor to retrieve.

        Returns:
            T: The sensor associated with the identifier.
        """
        return self.__values.__getitem__(item)

    def get(self, value: str, default: Optional[T] = None) -> Optional[T]:
        """Retrieves a sensor by identifier with a default if not found.

        Args:
            value (str): The sensor identifier.
            default (Optional[Any], optional): The default value if the sensor is not found.

        Returns:
            Optional[T]: The sensor or default value if not found.
        """
        return self.__values.get(value, default)

    def items(self) -> ItemsView[str, T]:
        """Returns an iterator over sensor items as (key, value) pairs.

        Returns:
            ItemsView[str, T]: An iterator over (identifier, sensor) pairs.
        """
        return self.__values.items()

    def values(self) -> ValuesView[T]:
        """Returns a view of all sensors.

        Returns:
            ValuesView[T]: A view of all sensors in the collection.
        """
        return self.__values.values()

    def keys(self) -> KeysView[str]:
        """Returns a view of all sensor identifiers.

        Returns:
            KeysView[str]: A view of all keys in the sensors dictionary.
        """
        return self.__values.keys()


class SensorDriver(Generic[T], SensorCallback[T], metaclass=ABCMeta):
    """A base class for a sensor driver with essential methods and callback handling.

    # Default functions
    For sensor attributes and attributes of classes that inherit from sensor, `setter`, `fetch`, and `refresh` functions
    are defined in the driver.

    The functions should follow the following convention:
    - they are named `fetch_[attribute_name]`, `refresh_[attribute_name]` and `set_[attribute_name]`
    - fetch functions:
        - Args: sensor (T): The sensor from which we gather the information.
        - Raise: SensorNotFound
        - Return: the data type as it is stored in the sensor (e.g., Union[TemperatureControl, NotAvailable])
    - refresh functions:
        - Args: sensor (T): The sensor from which we gather the information.
        - Raise: SensorNotFound
        - Return: None
        - are called in `refresh_attributes`.
    - setter functions
        - Args: sensor (T): The sensor from which we gather the information, the value as it is stored in the sensor.
                Receive the data type as it is stored in the sensor (e.g., Union[TemperatureControl, NotAvailable]) as second argument
                Receive the bool bounds_check: bool that indicates whether bounds should be checked.
        - Raise InputException
        - Return None
        - are called in `set_attributes`.
    """

    def __init__(self) -> None:
        """Initializes the SensorDriver with optional sensor tracking and a condition variable."""
        Generic.__init__(self)
        self.__sensors: Optional[Sensors[T]] = None
        self.__listeners: list[SensorCallback[T]] = []

    @property
    def sensors(self) -> Sensors[T]:
        """Gets the current sensors, raising an error if uninitialized.

        Returns:
            Sensors[T]: The current sensor collection.

        Raises:
            DriverInvalidState: If the sensors are uninitialized.

            DriverFeatureImplementationError: If the driver encountered a problem.
        """
        sensors = self.__sensors
        if sensors is None:
            raise DriverInvalidState()
        return sensors

    def _sensor(self, sensor: T) -> Optional[T]:
        """Retrieves a sensor instance by its identifier.

        Args:
            sensor (T): The sensor to retrieve, identified by its unique identifier.

        Returns:
            Optional[T]: The sensor instance if found.

        Raises:
            SensorNotFound: If the sensor identifier is not present in Sensors.
        """
        try:
            return self.sensors[sensor.identifier]
        except KeyError:
            raise SensorNotFound

    def destroy_sensors(self) -> None:
        """Destroys all sensors, setting their state to unavailable and handling exceptions.

        Raises:
            DriverInvalidState: If the sensors are uninitialized.

            DriverFeatureImplementationError: If the driver encountered a problem.
        """
        ex: Optional[DriverFeatureImplementationError] = None
        for sens in self.sensors.values():
            try:
                sens.state = SensorState.UNAVAILABLE
            except DriverFeatureImplementationError as e:
                ex = e

        self.__sensors = None
        if ex is not None:
            raise ex from None

    def create_sensors(self) -> None:
        """Creates a new sensor collection."""
        self.__sensors = Sensors[T](self)

    @abstractmethod
    def setup(self) -> 'SensorDriver[T]':
        """Initialize the driver. This function must be called before any other functionality can be used.

        Initializes the list of sensors and driver internals.

        Raises:
            DriverInvalidState if the driver is in an invalid state

            DriverFeatureImplementationError: If the driver encountered a problem.
        """

    @abstractmethod
    def destroy(self) -> None:
        """Destroys the driver. After calling destroy, no other function can be called successfully aside from setup.

        Destroys the list of sensors and unregisters driver internals.

        Raises:
            DriverInvalidState if the driver is in an invalid state

            DriverFeatureImplementationError: If the driver encountered a problem.
        """

    @abstractmethod
    def retrieve_sensors(self) -> None:
        """Retrieve all available sensors and update self.sensors.

        Does not refresh sensor attributes.

        Raises:
            DriverInvalidState if the driver is in an invalid state

            DriverFeatureImplementationError: If the driver encountered a problem.
        """

    def reset_driver(self) -> None:
        """Reset the driver.

        Temporarily disconnects and stops all cameras. This function can be called at any point in time.

        1. Driver internals are reset
        2. The driver stops and disconnects all sensors.
        3. Call to `destroy`
        4. Call to `setup`
        5. Call to `retrieve_sensors`
        6. The driver reconnects to all previously connected sensors and starts all previously started sensors.

        Raises:
            DriverInvalidState if the driver is in an invalid state

            SensorInvalidState if the sensor is in an invalid state

            DriverFeatureImplementationError if the driver feature is not implemented / ran into an error

            ResourceError: if the resource is currently not available.
        """
        raise DriverFeatureNotImplemented

    def __enter__(self) -> 'SensorDriver[T]':
        """Initialize the driver. This function must be called before any other functionality can be used.

        Initializes the list of sensors and driver internals.

        Returns:
            SensorDriver[T]: The setup driver instance.

        Raises:
            DriverInvalidState if the driver is in an invalid state

            DriverFeatureImplementationError: If the driver encountered a problem.

            DriverFeatureImplementationError: If an exception occurred in the implementation.
        """
        return self.setup()

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """Destroys the driver. After calling destroy, no other function can be called successfully aside from setup.

        Destroys the list of sensors and unregisters driver internals.

        Raises:
            DriverInvalidState if the driver is in an invalid state

            DriverFeatureImplementationError: If the driver encountered a problem.

            DriverFeatureImplementationError: If an exception occurred in the implementation.
        """
        self.destroy()

    #
    # Sensor state
    #

    @abstractmethod
    def connect(self, sensor: T) -> None:
        """Connect to the provided sensor.

        This creates the necessary internal connection objects.
        After the connection is established, the sensor's attributes are fetched and stored in the sensor instance.

        Arguments:
            sensor (T) The sensor which is supposed to be connected.

        Raises:
            InvalidState If the sensor or driver is in an invalid state

            DriverFeatureImplementationError: If an exception occurred in the implementation.

            ResourceError: If the requested resource is not available.

        See:
            `self.refresh_attributes`: Function used for gathering the attributes from the remote sensor.
        """

    @abstractmethod
    def disconnect(self, sensor: T) -> None:
        """Disconnect from the provided sensor.

        This removes previously created internal connection objects.

        Arguments:
            sensor (T) The sensor which is supposed to be disconnected.

        Raises:
            InvalidState: If the sensor or driver is in an invalid state

            DriverFeatureImplementationError: If an exception occurred in the implementation.

            ResourceError: If the requested resource is not available.
        """

    @abstractmethod
    def start(self, sensor: T) -> None:
        """Start the data acquisition from the provided sensor.

        Arguments:
            sensor (T): the camera that is supposed to be started.

        Raises:
            InvalidState: If the sensor or driver is in an invalid state

            FeatureUnavailable: If the feature is not available for the sensor or the driver.

            ResourceError: If the requested resource is not available.
        """

    @abstractmethod
    def stop(self, sensor: T) -> None:
        """Stop the data acquisition from the provided sensor.

        Arguments:
            sensor (T): the sensor that is supposed to be stopped.

        Raises:
            InvalidState: If the sensor or driver is in an invalid state

            FeatureUnavailable: If the feature is not available for the sensor or the driver.

            ResourceError: If the requested resource is not available.
        """

    def reset(self, sensor: T) -> None:
        """Reset the provided sensor.

        A reset
        1. stops image acquisition if image acquisition was started,
        2. performs a sensor reset,
        3. disconnects form the sensor,
        4. connects to the sensor, and
        5. optionally restarts the sensor if image acquisition was started.

        Arguments:
            sensor (T): the sensor that is supposed to be reset.

        Raises:
            InvalidState: If the sensor or driver is in an invalid state

            ResourceError: If the requested resource is not available.

            FeatureUnavailable: If the feature is not available for the sensor or the driver.
        """
        raise DriverFeatureNotImplemented

    def refresh_attributes(self, sensor: T) -> None:
        """Refresh all available attributes.

        Arguments:
            sensor (T): the sensor for which the attributes should be refreshed

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        sensor = self._sensor(sensor)
        refresh_functions: list[Callable[[T], None]] = [self.refresh_temperature_control]
        for fun in refresh_functions:
            try:
                fun(sensor)
            except SensorInvalidState as e:
                logger.debug(f'{fun.__name__} cannot be performed in the current sensor state {e}.')

            except (InvalidState, ConfigurationError) as e:
                logger.warning(f'{fun.__name__} raised an {type(e).__name__} error: {e}.')
                if is_application_debug():
                    traceback.print_exc()

            except ResourceError as e:
                logger.debug(f'{fun.__name__} raised a resource error {e}.')

            except SensorFeatureTemporarilyNotAvailable as e:
                logger.debug(f'{fun.__name__} temporarily not available {e}.')

            except (SensorFeatureUnavailable, DriverFeatureNotImplemented):
                pass

            except DriverFeatureImplementationError as e:
                logger.warning(f'{fun.__name__} implementation error{e}.')
                if is_application_debug():
                    traceback.print_exc()

            except InputException as e:
                logger.error(f'{fun.__name__} failed: {e}.')
                if is_application_debug():
                    traceback.print_exc()

    def set_attributes(self, sensor: T, bounds_check: bool = False) -> None:
        """Apply all attributes that were provided.

        Fails when the first exception is thrown.

        Arguments:
            sensor (T): a sensor configuration that should be applied, that has all the fields.

            bounds_check (bool):
                If true, adjusts the number values to adhere to the bounds required from the sensor.
                Otherwise, this function raises if we f.i. attempt to configure a gain value of 1000.0 even though the
                maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        old_configuration = self.sensors[sensor.identifier]

        setter = [
            (old_configuration.temperature_control, sensor.temperature_control, self.set_temperature_control),
        ]

        for old_config_value, new_config_value, func in setter:
            try:
                func(sensor, new_config_value, bounds_check=bounds_check)
            except InputException as e:
                if not old_config_value.entails(new_config_value):
                    raise e

    #
    # Sensor Configuration
    #

    def export_configuration(self, sensor: T) -> bytes:
        """Dump the configuration to a bytes array.

        Arguments:
            sensor (T) The sensor which is supposed to be disconnected.

        Return:
            bytes: Configuration dump.

        Raises:
            InvalidState: If the sensor or driver is in an invalid state

            ResourceError: If the requested resource is not available.

            FeatureUnavailable: If the feature is not available for the sensor or the driver.
        """
        raise DriverFeatureNotImplemented

    def import_configuration(self, sensor: T, configuration: bytes) -> None:
        """Read the configuration from a previous configuration dump.

        Arguments:
            sensor (T) The sensor which is supposed to be disconnected.

            configuration (bytes) : Configuration dump.

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    #
    # Sensor acquisition
    #

    @abstractmethod
    def execute_trigger(self, sensor: T, timestamp: Optional[int] = None) -> None:
        """Execute a trigger for acquiring data.

        Arguments:
            sensor (T):
                The camera that is about to be triggered.

            timestamp (Optional[int]):
                Trigger timestamp that can be used to identify the assign produced images to the trigger that caused
                them.

        Raises:
            InvalidState: If the sensor or driver is in an invalid state

            ResourceError: If the requested resource is not available.

            FeatureUnavailable: If the feature is not available for the sensor or the driver.
        """

    @abstractmethod
    def fetch_sensor_data(self, sensor: T, timeout: float = 0.0) -> Union[SensorData, NotAvailable]:
        """Fetches data from the specified sensor, waiting up to a defined timeout.

        Args:
            sensor (T): The sensor from which to retrieve data.

            timeout (float, optional): The maximum time in seconds to wait for data. Defaults to 0.0.

        Returns:
            Union[SensorData, NotAvailable]: The sensor data if available; otherwise, a NotAvailable status.

        Raises:
            InvalidState: If the sensor or driver is in an invalid state

            ResourceError: If the requested resource is not available.

            FeatureUnavailable: If the feature is not available for the sensor or the driver.
        """

    @abstractmethod
    def create_item(self, sensor: T, sensor_data: SensorData) -> Item:
        """Creates an item based on sensor data and associates it with the specified sensor.

        Args:
            sensor (T): The sensor for which the item is created.
            sensor_data (SensorData): The data used to create the item.

        Returns:
            Item: An item that contains all the relevant sensory information from the sensor data.

        Raises:
            ResourceError: If the requested resource is not available.
        """

    def ping(self) -> None:
        """If necessary, this can be implemented to check if the connection is still alive.
        This will be executed at a higher frequency, so don't do computationally expensive stuff here.

        Raise if a sensor disconnected.

        If this raises, the node will reset the connection.

        Raises:
            InvalidState: If the sensor or driver is in an invalid state

            ResourceError: If the requested resource is not available.

            FeatureUnavailable: If the feature is not available for the sensor or the driver.
        """

    #
    # Temperature control
    #

    def set_temperature_control(
        self, sensor: T, temperature_control: Union[TemperatureControl, NotAvailable], bounds_check: bool = False
    ) -> None:
        """Configure temperature control settings.

        Args:
            sensor (T): The sensor from which we gather the information.

            temperature_control (Union[TemperatureControl, NotAvailable]): the temperature control settings to apply.

            bounds_check (bool):
                If true, adjusts the number values to adhere to the bounds required from the sensor.
                Otherwise, this function raises if we f.i. attempt to configure a gain value of 1000.0 even though the
                maximum allowed gain is 100.

        Raises:
            InputException: If some settings could not be applied.
        """
        raise DriverFeatureNotImplemented

    def fetch_temperature_control(self, sensor: T) -> Union[TemperatureControl, NotAvailable]:  # noqa
        """Retrieve the temperature control settings from camera.

        Args:
            sensor (T): The sensor from which we gather the information.

        Returns:
            Union[TemperatureControl, NotAvailable] the temperature control settings.

        Raises:
            SensorNotFound if the sensor is currently not found.
        """
        return NotAvailable()

    def refresh_temperature_control(self, sensor: T) -> None:
        """Fetch and update the temperature control settings.

        Arguments:
            sensor (T): The sensor from which we gather the information.

        Raises:
            InvalidState: If the sensor or driver is in an invalid state

            ResourceError: If the requested resource is not available.

            FeatureUnavailable: If the feature is not available for the sensor or the driver.
        """
        sensor = self._sensor(sensor)
        sensor.temperature_control = self.fetch_temperature_control(sensor)

    #
    # Callbacks
    #

    def add_listener(self, listener: SensorCallback[T]) -> None:
        """Adds a listener for sensor events.

        Args:
            listener (SensorCallback[T]): The listener to be added for handling sensor events.
                                          Listeners are only expected to throw input exceptions.
        """
        self.__listeners.append(listener)

    def on_sensor_tracked(self, sensor_identifier: str) -> None:
        """Notifies listeners when a sensor is tracked.

        Args:
            sensor_identifier (str): The identifier of the tracked sensor.

        Raises:
            DriverFeatureImplementationError: If a listener encounters an error during execution.
        """
        for listener in self.__listeners:
            try:
                listener.on_sensor_tracked(sensor_identifier)
            except DriverFeatureImplementationError as e:
                raise e from None
            except Exception as e:
                raise DriverFeatureImplementationError(
                    f'Callback raises invalid exception: {type(e).__name__}: {e}'
                ) from e

    def on_sensor_found(self, sensor: T) -> None:
        """Notifies listeners when a sensor is found.

        Args:
            sensor (T): The sensor that has been found.

        Raises:
            DriverFeatureImplementationError: If a listener encounters an error during execution.
        """
        for listener in self.__listeners:
            try:
                listener.on_sensor_found(sensor)
            except DriverFeatureImplementationError as e:
                raise e from None
            except Exception as e:
                raise DriverFeatureImplementationError(
                    f'Callback raises invalid exception: {type(e).__name__}: {e}'
                ) from e

    def on_sensor_lost(self, sensor: T) -> None:
        """Notifies listeners when a sensor is lost.

        Args:
            sensor (T): The sensor that has been lost.

        Raises:
            DriverFeatureImplementationError: If a listener encounters an error during execution.
        """
        for listener in self.__listeners:
            try:
                listener.on_sensor_lost(sensor)
            except DriverFeatureImplementationError as e:
                raise e from None
            except Exception as e:
                raise DriverFeatureImplementationError(
                    f'Callback raises invalid exception: {type(e).__name__}: {e}'
                ) from e

    def on_sensor_connected(self, sensor: T) -> None:
        """Notifies listeners when a sensor is connected.

        Args:
            sensor (T): The sensor that has been connected.

        Raises:
            DriverFeatureImplementationError: If a listener encounters an error during execution.
        """
        for listener in self.__listeners:
            try:
                listener.on_sensor_connected(sensor)
            except DriverFeatureImplementationError as e:
                raise e from None
            except Exception as e:
                raise DriverFeatureImplementationError(
                    f'Callback raises invalid exception: {type(e).__name__}: {e}'
                ) from e

    def on_sensor_disconnected(self, sensor: T) -> None:
        """Notifies listeners when a sensor is disconnected.

        Args:
            sensor (T): The sensor that has been disconnected.

        Raises:
            DriverFeatureImplementationError: If a listener encounters an error during execution.
        """
        for listener in self.__listeners:
            try:
                listener.on_sensor_disconnected(sensor)
            except DriverFeatureImplementationError as e:
                raise e from None
            except Exception as e:
                raise DriverFeatureImplementationError(
                    f'Callback raises invalid exception: {type(e).__name__}: {e}'
                ) from e

    def on_sensor_started(self, sensor: T) -> None:
        """Notifies listeners when a sensor has started.

        Args:
            sensor (T): The sensor that has started.

        Raises:
            DriverFeatureImplementationError: If a listener encounters an error during execution.
        """
        for listener in self.__listeners:
            try:
                listener.on_sensor_started(sensor)
            except DriverFeatureImplementationError as e:
                raise e from None
            except Exception as e:
                raise DriverFeatureImplementationError(
                    f'Callback raises invalid exception: {type(e).__name__}: {e}'
                ) from e

    def on_sensor_stopped(self, sensor: T) -> None:
        """Notifies listeners when a sensor has stopped.

        Args:
            sensor (T): The sensor that has stopped.

        Raises:
            DriverFeatureImplementationError: If a listener encounters an error during execution.
        """
        for listener in self.__listeners:
            try:
                listener.on_sensor_stopped(sensor)
            except DriverFeatureImplementationError as e:
                raise e from None
            except Exception as e:
                raise DriverFeatureImplementationError(
                    f'Callback raises invalid exception: {type(e).__name__}: {e}'
                ) from e

    def on_sensor_updated(self, sensor: T) -> None:
        """Notifies listeners when a sensor is updated.

        Args:
            sensor (T): The sensor that has been updated.

        Raises:
            DriverFeatureImplementationError: If a listener encounters an error during execution.
        """
        for listener in self.__listeners:
            try:
                listener.on_sensor_updated(sensor)
            except DriverFeatureImplementationError as e:
                raise e from None
            except Exception as e:
                raise DriverFeatureImplementationError(
                    f'Callback raises invalid exception: {type(e).__name__}: {e}'
                ) from e
