from abc import ABCMeta
import time
import traceback
import copy
import numpy as np
import logging
import os
from dataclasses import dataclass, field
from functools import singledispatchmethod, partial

from typing import Generic, TypeVar, Any, Callable, Optional
from queue import Queue, Empty

from genicam.genapi import AccessException, InvalidArgumentException

from dataspree.inference.core.config import ConfigFieldTypes, bool_input
from dataspree.inference.core.producer import Producer, PropertyType, Property
from dataspree.inference.core.decorators import action, Item
from dataspree.inference.core.signals import BaseSignal
from dataspree.inference.core.utils import inference_config_subdir, get_expanded_path
from dataspree.inference.sensor.data import Sensor, SensorState, Camera
from dataspree.inference.sensor.data.camera import TriggerAcquisitionMode
from dataspree.inference.sensor.data.exceptions import (
    SensorException,
    DriverInvalidState,
    FeatureUnavailable,
    SensorLost,
)
from dataspree.inference.sensor.data.sensor import SensorStateTransition
from dataspree.inference.sensor.data.sensor_callback import SensorCallback
from dataspree.inference.sensor.data.configuration import is_available
from dataspree.inference.sensor.driver import SensorDriver
from dataspree.inference.core.node import Node

from dataspree.inference.core.environment import is_application_debug
from dataspree.inference.sensor.signals import (
    RequestFailedSignal,
    SensorApplySettingsRequestSignal,
    SensorTriggerRequestSignal,
    SensorRequestReconnectSignal,
    SensorRequestDisconnectSignal,
    SensorTriggerRequestSignalFailed,
    SensorRequestReconnectFailedSignal,
    SensorRequestDisconnectFailedSignal,
    SensorApplySettingsRequestFailedSignal,
    DriverConnectedSignal,
    DriverDisconnectedSignal,
    SensorStateTransitionSignal,
    DriverRequestReset,
    DriverRequestResetFailed,
)

J = TypeVar('J', bound=Sensor)
T = TypeVar('T', bound=SensorDriver)


@dataclass
class TemporaryInformationSensor:
    producer: Producer
    image_counter: int = 0
    frame_times: np.ndarray = field(default_factory=lambda: np.empty(50, dtype=float))

    def tic(self):
        self.frame_times[self.image_counter % len(self.frame_times)] = time.time()
        self.image_counter += 1

    @property
    def fps(self) -> float:
        newest = (self.image_counter - 1 + len(self.frame_times)) % len(self.frame_times)
        oldest = self.image_counter % len(self.frame_times)

        length = min(len(self.frame_times), self.image_counter)
        if self.image_counter < len(self.frame_times):
            oldest = 0

        fps = (
            length / (self.frame_times[newest] - self.frame_times[oldest])
            if (self.frame_times[oldest] != self.frame_times[newest])
            else 0.0
        )
        return fps


U = TypeVar('U', bound=TemporaryInformationSensor)


class SensorInput(Node, Generic[J, T, U], SensorCallback[J], metaclass=ABCMeta):
    @dataclass
    class QueueableActions:
        function: Callable[[T], None]
        signal_success: Optional[RequestFailedSignal] = None
        signal_error: Optional[RequestFailedSignal] = None
        action_success: Optional['QueueableAction'] = None
        action_error: Optional['QueueableAction'] = None

    def __init__(
        self,
        name: str,
        driver: T,
        producer: Producer,
        required_sensors: Optional[list[str]] = None,
        delay: float = 0.0,
        configuration_root_dir: Optional[str] = None,
        auto_connect_all: bool = True,
    ) -> None:
        Node.__init__(self, name, True, True)
        Generic.__init__(self)
        SensorCallback.__init__(self)

        self.producer = producer
        self.add_producer(self.producer)

        self.driver: T = driver
        self.required_sensors: list[str] = required_sensors or []
        self.configuration_root_dir: Optional[str] = configuration_root_dir
        self.auto_connect_all: bool = auto_connect_all

        self.__sensor_mapping: dict[str, U] = {}
        self.__execution_queue: Queue[SensorInput[J, T, U].QueueableActions] = Queue()
        self.paused: bool = False
        self.delay: float = delay

        driver.add_listener(self)

        self.__total_image_count: int = 0
        self.__camera_image_count: dict[str, int] = {}
        self.__previous_log_level: int = self.logger.level

    @property
    def sensor_configuration_root_dir(self) -> str:
        if self.configuration_root_dir:
            return self.configuration_root_dir
        return inference_config_subdir('sensors')

    def get_sensor_configuration_file(self, sensor: J):
        return get_expanded_path(
            os.path.join(self.sensor_configuration_root_dir, f'{self.name}_{sensor.identifier}.yml')
        )

    @classmethod
    def create_producer(cls, name: str) -> Producer:
        return Producer(name, [Property('image', PropertyType.IMAGE)])

    def __transmit_signal(self, response_signal: Optional[RequestFailedSignal], e: Optional[Exception]):
        if response_signal is None:
            return

        if e is not None:
            response_signal.set_reason(e)
        self.propagate_signal(response_signal)

    def run(self) -> None:
        while self.running:
            try:
                self.__execution_queue: Queue[SensorInput[J, T, U].QueueableActions] = Queue()

                with self.driver as connection:
                    self.propagate_signal(DriverConnectedSignal(self))
                    connection.retrieve_sensors()

                    while self.running:
                        # TODO: We need to do this more often (if paused, this would limit the signal throughput)
                        self.direct_signal_consumer.receive(True, self.delay)
                        # connection.ping()

                        try:
                            missing_sensors = self.get_missing_sensors(connection)
                            sensor_missing = len(missing_sensors)
                            if sensor_missing:
                                self.logger.debug(f'Searching for missing sensors' f' {", ".join(missing_sensors)}.')

                            if sensor_missing or not any(
                                f for f in connection.sensors.values() if f.state != SensorState.UNAVAILABLE
                            ):
                                self.sensors_retrieve()

                            while self.running:
                                try:
                                    root_el = self.__execution_queue.get(block=False)
                                except Empty:
                                    raise Empty

                                el = root_el

                                while el is not None:
                                    exception: Optional[Exception] = None
                                    try:
                                        el.function(connection)

                                    except DriverInvalidState as e:
                                        raise e from None

                                    except ReferenceError as e:
                                        raise e from None

                                    except FeatureUnavailable as e:
                                        exception = e
                                        self.logger.warning(f'The requested feature is (currently not available: {e}')
                                        traceback.print_exc()

                                    except SensorLost as e:
                                        e.sensor.state = SensorState.UNAVAILABLE
                                        self.sensors_retrieve()
                                        traceback.print_exc()

                                    except SensorException as e:
                                        exception = e
                                        self.logger.warning(f'A sensor exception occurred: {e}')

                                    except InvalidArgumentException as e:
                                        exception = e
                                        self.logger.warning(f'Invalid argument {e}')
                                        if is_application_debug():
                                            traceback.print_exc()

                                    except Exception as e:
                                        exception = e
                                        self.logger.error(
                                            f'An unexpected exception of type "'
                                            f'{type(e).__name__}" '
                                            f'occurred: {e}'
                                        )

                                        if is_application_debug():
                                            traceback.print_exc()

                                        time.sleep(0.01)

                                    finally:
                                        success: bool = exception is None
                                        if success:
                                            self.__transmit_signal(el.signal_success, exception)
                                            el = el.action_success
                                        else:
                                            self.__transmit_signal(el.signal_error, exception)
                                            el = el.action_error
                        except Empty:
                            pass

                        if self.paused:
                            time.sleep(0.05)
                            continue

                        for sensor in connection.sensors.values():
                            ti: Optional[U] = self.__sensor_mapping.get(sensor.identifier)
                            if ti is not None:
                                producer = ti.producer
                                if self.before_image_acquire(sensor, connection, ti):
                                    fetched_buffer = connection.fetch_sensor_data(sensor, timeout=0.00001)
                                    if is_available(fetched_buffer):
                                        item: Optional[Item] = connection.create_item(sensor, fetched_buffer)
                                        if item is not None:
                                            ti.tic()
                                            adjusted_item: Optional[Item] = self.on_item_acquired(
                                                sensor, connection, ti, item
                                            )
                                            if adjusted_item is not None:
                                                producer.send(adjusted_item)
                                                self.producer.send(adjusted_item)
                                                producer.send(
                                                    Item.default(display_image=adjusted_item['image'][..., 0])
                                                )

            except SensorException:
                reconnect_delay_seconds: int = 1
                self.logger.info(f'Sensor lost. Reconnecting in {reconnect_delay_seconds}s')
                time.sleep(reconnect_delay_seconds)

            except DriverInvalidState:
                reconnect_delay_seconds: int = 1
                self.logger.warning(f'A driver state exception occurred. Reconnecting in {reconnect_delay_seconds}s')
                time.sleep(reconnect_delay_seconds)

            except ReferenceError:
                reconnect_delay_seconds: int = 1
                self.logger.info(f'Harvester disconnected. Reconnecting in {reconnect_delay_seconds}s')
                time.sleep(reconnect_delay_seconds)

            except Exception as e:
                reconnect_delay_seconds: int = 1
                self.logger.warning(
                    f'An exception occurred while connecting to the camera: {e}. '
                    f'Reconnecting in {reconnect_delay_seconds}s'
                )

                if is_application_debug():
                    traceback.print_exc()
                time.sleep(reconnect_delay_seconds)

                self.propagate_signal(DriverDisconnectedSignal(self, e))

            finally:
                self.propagate_signal(DriverDisconnectedSignal(self))

        self.logger.info('Shut down!')

    @action
    def verbose(self):
        self.logger.setLevel(logging.DEBUG)

    @action
    def verbose_off(self):
        self.logger.setLevel(self.__previous_log_level)

    #
    # Functions that can be implemented by specializations.
    #

    def before_image_acquire(self, sensor: J, driver: T, temporary_information: U) -> bool:
        """TODO: Write documentation
        TODO: Find out if there is a specific decorator like abstractmethod that does
            - not force you to implement it (we use the default impl instead (return True))
            - does not complain about unused args.

        """
        return True

    def on_item_acquired(self, sensor: J, driver: T, temporary_information: U, item: Item) -> Optional[Item]:
        """TODO: Write documentation."""
        item['frame_count'] = self.__sensor_mapping[sensor.name].image_counter
        return item

    def create_sensor_representation(self, sensor: J) -> U:
        return TemporaryInformationSensor(producer=self.create_producer(f'{self.name}_{sensor.identifier}'))

    # Discover
    def on_sensor_tracked(self, sensor_identifier: str) -> None:
        self.logger.debug(f'Sensor "{sensor_identifier}" tracked.')

    def get_missing_sensors(self, driver: T) -> list[str]:
        driver_sensor_identifiers = [k for k, v in driver.sensors.items() if v.state != SensorState.UNAVAILABLE]

        return [s for s in self.required_sensors if s not in driver_sensor_identifiers]

    def on_sensor_found(self, sensor: J) -> None:
        self.logger.info(f'Sensor "{sensor.identifier}" discovered.')
        self.propagate_signal(SensorStateTransitionSignal(self, sensor, SensorStateTransition.FOUND))

        if self.auto_connect_all or sensor.identifier in self.required_sensors:
            action_1 = partial(type(self.driver).connect, sensor=sensor)
            action_2 = partial(type(self.driver).start, sensor=sensor)
            self.schedule_main_loop_execution(
                action_1, None, action_success=SensorInput[J, T, U].QueueableActions(action_2, None)
            )

    def on_sensor_lost(self, sensor: J) -> None:
        self.logger.warning(f'Sensor "{sensor.identifier}" lost.')
        self.propagate_signal(SensorStateTransitionSignal(self, sensor, SensorStateTransition.LOST))

        info = self.__sensor_mapping.pop(sensor.identifier, None)
        if info is not None:
            self.remove_producer(info.producer)

    def on_sensor_connected(self, sensor: J) -> None:
        self.logger.info(f'Sensor "{sensor.identifier}" connected.')
        self.propagate_signal(SensorStateTransitionSignal(self, sensor, SensorStateTransition.CONNECTED))

    def on_sensor_disconnected(self, sensor: J) -> None:
        self.logger.info(f'Sensor "{sensor.identifier}" disconnected.')
        self.propagate_signal(SensorStateTransitionSignal(self, sensor, SensorStateTransition.DISCONNECTED))

    def on_sensor_started(self, sensor: J) -> None:
        self.logger.info(f'Sensor "{sensor.identifier}" started.')
        self.propagate_signal(SensorStateTransitionSignal(self, sensor, SensorStateTransition.STARTED))

        if sensor.identifier in self.__sensor_mapping:
            self.logger.warning(f'Producer for sensor {sensor.identifier} was not removed.')
            self.remove_producer(self.__sensor_mapping[sensor.identifier].producer)

        self.__sensor_mapping[sensor.identifier] = self.create_sensor_representation(sensor)
        self.add_producer(self.__sensor_mapping[sensor.identifier].producer)

    def on_sensor_stopped(self, sensor: J) -> None:
        self.logger.info(f'Sensor "{sensor.identifier}" stopped.')
        self.propagate_signal(SensorStateTransitionSignal(self, sensor, SensorStateTransition.STOPPED))

        info = self.__sensor_mapping.pop(sensor.identifier, None)
        if info is not None:
            self.remove_producer(info.producer)
        else:
            self.logger.warning(f'Producer for {sensor.identifier} already removed')
        # TODO: Remove the producer

    def on_sensor_updated(self, sensor: T) -> None:
        self.logger.debug(f'Sensor "{sensor}" was updated.')

    #
    # API
    #

    def schedule_main_loop_execution(
        self, execution: Callable[[T], None], signal_error, signal_success=None, action_error=None, action_success=None
    ) -> None:
        self.__execution_queue.put(
            SensorInput[J, T, U].QueueableActions(
                function=execution,
                signal_error=signal_error,
                signal_success=signal_success,
                action_error=action_error,
                action_success=action_success,
            )
        )

    #
    # Interactions
    #
    @action
    def driver_reset(self, on_error=None):
        self.schedule_main_loop_execution(type(self.driver).reset_driver, on_error)

    @action
    def sensors_retrieve(self, on_error=None) -> None:
        self.schedule_main_loop_execution(type(self.driver).retrieve_sensors, on_error)

    @action
    def sensor_restart(self, exclude_disconnected: bool = False, exclude_stopped: bool = False, on_error=None):
        def reconnect_all_sensors(connection: T) -> None:
            for sensor in connection.sensors.values():
                try:
                    was_stopped = sensor.state != SensorState.STARTED
                    was_disconnected = was_stopped and sensor.state != SensorState.CONNECTED

                    if was_stopped and exclude_stopped:
                        continue

                    if was_disconnected and exclude_disconnected:
                        continue

                    if sensor.state == SensorState.STARTED:
                        connection.stop(sensor)

                    if sensor.state == SensorState.CONNECTED:
                        connection.disconnect(sensor)

                    if sensor.state == SensorState.DISCOVERED:
                        connection.connect(sensor)

                    if sensor.state == SensorState.CONNECTED:
                        connection.start(sensor)
                except SensorException as e:
                    self.logger.warning(f'Error reconnecting {sensor.identifier}: {e}')

        self.schedule_main_loop_execution(reconnect_all_sensors, on_error)

    def restart_sensor(self, sensor: Optional[J] = None, on_error=None):
        def restart_sensor(connection: T) -> None:
            sensors: list[J] = []
            if sensor is None:
                sensors = list(connection.sensors.values())
            else:
                if sensor.identifier is None:
                    for sid in list(connection.sensors.keys()):
                        current_sensor = copy.copy(sensor)
                        current_sensor.identifier = sid
                        sensors.append(current_sensor)
                else:
                    sensors = list(connection.sensors.values())

            for current_sensor in sensors:
                refreshed_sensor = connection.sensors[current_sensor.identifier]

                if refreshed_sensor.state == SensorState.STARTED:
                    connection.stop(refreshed_sensor)

                if refreshed_sensor.state == SensorState.CONNECTED:
                    connection.disconnect(refreshed_sensor)

                if refreshed_sensor.state == SensorState.DISCOVERED:
                    connection.connect(refreshed_sensor)

                if refreshed_sensor.state == SensorState.CONNECTED:
                    connection.start(refreshed_sensor)

        self.schedule_main_loop_execution(restart_sensor, on_error)

    def safe_disconnect_sensor(self, sensor: Optional[J], on_error=None):
        def safe_disconnect_sensor(connection: T) -> None:
            sensors: list[J] = []
            if sensor is None:
                sensors = list(connection.sensors.values())
            else:
                if sensor.identifier is None:
                    for sid in list(connection.sensors.keys()):
                        current_sensor = copy.copy(sensor)
                        current_sensor.identifier = sid
                        sensors.append(current_sensor)
                else:
                    sensors = list(connection.sensors.values())

            for current_sensor in sensors:
                refreshed_sensor = connection.sensors[current_sensor.identifier]

                if refreshed_sensor.state == SensorState.STARTED:
                    try:
                        connection.stop(refreshed_sensor)
                    except SensorLost:
                        pass

                if refreshed_sensor.state == SensorState.CONNECTED:
                    try:
                        connection.disconnect(refreshed_sensor)
                    except SensorLost:
                        pass

        self.schedule_main_loop_execution(safe_disconnect_sensor, on_error)

    @action
    def sensor_connect(self, on_error=None):
        def connect_sensors(connection: T) -> None:
            for sensor in connection.sensors.values():
                try:
                    connection.connect(sensor)
                except SensorException as e:
                    self.logger.warning(f'Error connecting {sensor.identifier}: {e}')

        self.schedule_main_loop_execution(connect_sensors, on_error)

    def connect_sensor(self, sensor: J, on_error=None):
        self.schedule_main_loop_execution(partial(type(self.driver).connect, sensor=sensor), on_error)

    @action
    def sensor_disconnect(self, on_error=None):
        def connect_sensors(connection: T) -> None:
            for sensor in connection.sensors.values():
                try:
                    connection.disconnect(sensor)
                except SensorException as e:
                    self.logger.warning(f'Error disconnecting {sensor.identifier}: {e}')

        self.schedule_main_loop_execution(connect_sensors, on_error)

    def disconnect_sensor(self, sensor: J, on_error=None):
        self.schedule_main_loop_execution(partial(type(self.driver).disconnect, sensor=sensor), on_error)

    @action
    def sensor_start(self, on_error=None):
        def start_sensors(connection: T) -> None:
            for sensor in connection.sensors.values():
                try:
                    connection.start(sensor)
                except SensorException as e:
                    self.logger.warning(f'Error starting {sensor.identifier}: {e}')

        self.schedule_main_loop_execution(start_sensors, on_error)

    def start_sensor(self, sensor: J, on_error=None):
        self.schedule_main_loop_execution(partial(type(self.driver).start, sensor=sensor), on_error)

    @action
    def sensor_stop(self, on_error=None):
        def stop_sensors(connection: T) -> None:
            for sensor in connection.sensors.values():
                try:
                    connection.stop(sensor)
                except SensorException as e:
                    self.logger.warning(f'Error starting {sensor.identifier}: {e}')

        self.schedule_main_loop_execution(stop_sensors, on_error)

    def stop_sensor(self, sensor: J, on_error=None):
        self.schedule_main_loop_execution(partial(type(self.driver).stop, sensor=sensor), on_error)

    @action
    def sensor_reset(self, on_error=None):
        def reset_sensors(connection: T) -> None:
            for sensor in connection.sensors.values():
                try:
                    connection.reset(sensor)
                except SensorException as e:
                    self.logger.warning(f'Error resetting {sensor.identifier}: {e}')
                except AccessException as e:
                    self.logger.warning(f'Error resetting {sensor.identifier}: access denied {e}')

        self.schedule_main_loop_execution(reset_sensors, on_error)

    def reset_sensor(self, sensor: J, on_error=None):
        self.schedule_main_loop_execution(partial(type(self.driver).reset, sensor=sensor), on_error)

    @action
    def sensors_refresh_attributes(self, on_error=None):
        def refresh_sensors_attributes(connection: T) -> None:
            for sensor in connection.sensors.values():
                try:
                    connection.refresh_attributes(sensor)
                except SensorException as e:
                    self.logger.warning(f'Error refreshing sensor attributes for {sensor.identifier}: {e}')
                except AccessException as e:
                    self.logger.warning(
                        f'Error refreshing sensor attributes for {sensor.identifier}: ' f'access denied {e}'
                    )

        self.schedule_main_loop_execution(refresh_sensors_attributes, on_error)

    def refresh_sensor_attributes(self, sensor: J, on_error=None):
        self.schedule_main_loop_execution(partial(type(self.driver).refresh_attributes, sensor=sensor), on_error)

    @action
    def sensor_export_configurations(self, on_error=None) -> None:
        def export_configurations(connection: T) -> None:
            for sensor in connection.sensors.values():
                try:
                    output = connection.export_configuration(sensor)
                    with open(self.get_sensor_configuration_file(sensor), 'wb') as file:
                        file.write(output)

                    self.logger.info(f'Exported the configuration to {self.get_sensor_configuration_file(sensor)}.')
                except SensorException as e:
                    self.logger.warning(f'Error refreshing sensor attributes for {sensor.identifier}: {e}')

        self.schedule_main_loop_execution(export_configurations, on_error)

    def export_sensor_configuration(self, sensor: J, output_file: Optional[str] = None, on_error=None):
        def export(connection: T) -> None:
            output = connection.export(sensor)
            with open(output_file, 'wb') as file:
                file.write(output)

        self.schedule_main_loop_execution(export, on_error)

    @action
    def sensor_import_configurations(self, on_error=None) -> None:
        def import_configurations(connection: T) -> None:
            for sensor in connection.sensors.values():
                try:
                    with open(self.get_sensor_configuration_file(sensor), mode='rb') as file:
                        configuration = file.read()
                    connection.import_configuration(sensor=sensor, configuration=configuration)
                except SensorException as e:
                    self.logger.warning(f'Error refreshing sensor attributes for {sensor.identifier}: {e}')

        self.schedule_main_loop_execution(import_configurations, on_error)

    def import_sensor_configuration(self, sensor: J, input_file: Optional[str] = None, on_error=None):
        with open(input_file, mode='rb') as file:
            configuration = file.read()
        self.schedule_main_loop_execution(
            partial(type(self.driver).import_configuration(sensor=sensor, configuration=configuration)), on_error
        )

    @action
    def sensor_trigger(self, sensor: Optional[J] = None, on_error=None, timestamp=None, adjust_acquisition_mode=False):
        def trigger_sensors(connection: T) -> None:
            sensors: list[J] = []
            if sensor is None:
                sensors = list(connection.sensors.values())
            else:
                if not is_available(sensor.identifier):
                    for sid, other_sensor in connection.sensors.items():
                        if other_sensor.state != SensorState.STARTED:
                            continue

                        current_sensor = copy.deepcopy(sensor)
                        current_sensor.identifier = sid
                        sensors.append(current_sensor)
                else:
                    sensors = list(connection.sensors.values())

            for current_sensor in sensors:
                try:
                    connection.set_attributes(current_sensor)
                    if adjust_acquisition_mode:
                        if isinstance(current_sensor, Camera) and not isinstance(
                            current_sensor.acquisition_mode, TriggerAcquisitionMode
                        ):
                            connection.configure_acquisition_trigger_based(current_sensor)

                    connection.execute_trigger(current_sensor, timestamp)
                except SensorException as e:
                    self.logger.warning(f'Error triggering {current_sensor.identifier}: {e}')
                    raise e from None
                except SensorLost as e:
                    raise e from None

        self.schedule_main_loop_execution(trigger_sensors, on_error)

    def apply_configuration(self, new_configuration: Optional[T], on_error=None):
        self.schedule_main_loop_execution(partial(type(self.driver).set_attributes, sensor=new_configuration), on_error)

    @action
    def pause(self) -> None:
        self.paused = True

    @action
    def resume(self) -> None:
        self.paused = False

    #
    # Signals
    #

    @singledispatchmethod
    def process_signal(self, signal: BaseSignal) -> None:
        self.logger.info(f'{self} received {signal}')
        super().process_signal(signal)

    @process_signal.register
    def process_apply_settings_signal(self, signal: SensorApplySettingsRequestSignal):
        self.logger.info(f'{self} received apply settings request {signal}')
        self.apply_configuration(
            signal.sensor,
            SensorApplySettingsRequestFailedSignal(
                self,
                signal,
            ),
        )

    @process_signal.register
    def process_trigger_request(self, signal: SensorTriggerRequestSignal):
        self.logger.info(f'{self} received sensor trigger request {signal}')
        self.sensor_trigger(
            signal.sensor,
            SensorTriggerRequestSignalFailed(self, signal),
            timestamp=signal.timestamp,
            adjust_acquisition_mode=signal.adjust_acquisition_mode,
        )

    @process_signal.register
    def process_restart_sensor(self, signal: SensorRequestReconnectSignal):
        self.logger.info(f'{self} received sensor restart request {signal}')
        self.restart_sensor(signal.sensor, SensorRequestReconnectFailedSignal(self, signal))

    @process_signal.register
    def process_disconnect_sensor(self, signal: SensorRequestDisconnectSignal):
        self.logger.info(f'{self} received sensor disconnect request {signal}')
        self.safe_disconnect_sensor(signal.sensor, SensorRequestDisconnectFailedSignal(self, signal))

    @process_signal.register
    def process_reset_driver(self, signal: DriverRequestReset):
        self.logger.info(f'{self} driver reset request {signal}')
        self.driver_reset(DriverRequestResetFailed(self, signal))

    #
    # Configuration
    #

    @classmethod
    def get_config_entries(cls) -> dict[str, Any]:
        return {
            'display_name': 'Sensor input',
            'description': '',
            'fields': [
                {
                    'name': 'delay',
                    'type': ConfigFieldTypes.DOUBLE_INPUT,
                    'display_name': 'delay',
                    'description': 'Delay for execution loop.',
                },
                {
                    'name': 'configuration_root_dir',
                    'type': ConfigFieldTypes.STRING_INPUT,
                    'display_name': 'Sensor configuration root folder',
                    'default': '',
                },
                bool_input('auto_connect_all'),
            ],
        }

    @staticmethod
    def get_config_categories() -> list[str]:
        return ['inputs']
