"""This module defines various signals related to sensor operations.

Signals include driver state changes, sensor state transitions, and requests for sensor actions such as applying
settings, triggering, reconnecting, and disconnecting.
"""

from typing import Optional, TypeVar, Generic
from logging import getLogger


from dataspree.inference.core.signals import NodeSignal
from dataspree.inference.core.node import Node
from dataspree.inference.sensor.data import Sensor, Camera
from dataspree.inference.sensor.data.sensor import SensorStateTransition

logger = getLogger(__name__)


class RequestFailedSignal(NodeSignal):
    """Signal that is ment not to be instantiated, but provides a unique interface to all the failed signals."""

    def __init__(self, source: Node, reason: Optional[Exception] = None) -> None:
        """Signal that is meant not to be instantiated, but provides a unique interface to all the failed signals.

        Args:
            source (Node): The node that initiated the signal.
            reason (Optional[Exception]): The reason for the failure, if any.
        """
        NodeSignal.__init__(self, source)
        self._reason: Optional[Exception] = reason

    @property
    def reason(self) -> Optional[Exception]:
        """Return the reason."""
        return self._reason

    @reason.setter
    def reason(self, reason: Optional[Exception] = None) -> None:
        """Set the reason."""
        self._reason = reason


#
# Driver state
#


class DriverConnectedSignal(NodeSignal):
    """Signal indicating that the driver has connected."""

    def __init__(self, source: Node) -> None:
        """Signal indicating that the driver has connected.

        Args:
            source (Node): The node that initiated the signal.
        """
        NodeSignal.__init__(self, source)


class DriverDisconnectedSignal(NodeSignal):
    """Signal indicating that the driver has disconnected."""

    def __init__(self, source: Node, reason: Optional[Exception] = None) -> None:
        """Signal indicating that the driver has disconnected.

        Args:
            source (Node): The node that initiated the signal.
            reason (Optional[Exception]): The reason for disconnection, if any.
        """
        NodeSignal.__init__(self, source)
        self.reason: Optional[Exception] = reason


#
# Request driver reset
#


class DriverRequestReset(NodeSignal):
    """Signal requesting a driver reset."""

    def __init__(self, source: Node) -> None:
        """Signal requesting a driver reset.

        Args:
            source (Node): The node that initiated the signal.
        """
        NodeSignal.__init__(self, source)


class DriverRequestResetFailed(RequestFailedSignal):
    """Signal indicating that a driver reset request has failed."""

    def __init__(self, source: Node, request: DriverRequestReset, reason: Optional[Exception] = None) -> None:
        """Signal indicating that a driver reset request has failed.

        Args:
            source (Node): The node that initiated the signal.
            request (DriverRequestReset): The reset request that failed.
            reason (Optional[Exception]): The reason for the failure, if any.
        """
        RequestFailedSignal.__init__(self, source, reason=reason)
        self.request: SensorRequestReconnectSignal[T] = request


T = TypeVar('T', bound=Sensor)


#
# Report state transition
#


class SensorStateTransitionSignal(NodeSignal, Generic[T]):
    """Signal indicating a sensor state transition."""

    def __init__(self, source: Node, sensor: T, target_state: SensorStateTransition) -> None:
        """Signal indicating a sensor state transition.

        Args:
            source (Node): The node that initiated the signal.
            sensor (T): The sensor undergoing the state transition.
            target_state (SensorStateTransition): The target state of the transition.
        """
        NodeSignal.__init__(self, source)
        self.sensor: T = sensor
        self.target_state_transition: SensorStateTransition = target_state


class CameraStateTransitionSignal(SensorStateTransitionSignal[Camera]):
    """Signal indicating a camera state transition."""


#
# Request settings application
#


class SensorApplySettingsRequestSignal(NodeSignal, Generic[T]):
    """Signal requesting the application of sensor settings."""

    def __init__(self, source: Node, settings: T) -> None:
        """Signal requesting the application of sensor settings.

        Args:
            source (Node): The node that initiated the signal.
            settings (T): The settings to be applied to the sensor.
        """
        NodeSignal.__init__(self, source)
        self.settings: T = settings


class CameraApplySettingsRequestSignal(SensorApplySettingsRequestSignal[Camera]):
    """Signal requesting the application of camera settings."""


class SensorApplySettingsRequestFailedSignal(RequestFailedSignal, Generic[T]):
    """Signal indicating that a request to apply sensor settings has failed."""

    def __init__(
        self, source: Node, request: SensorApplySettingsRequestSignal[T], reason: Optional[Exception] = None
    ) -> None:
        """Signal indicating that a request to apply sensor settings has failed.

        Args:
            source (Node): The node that initiated the signal.
            request (SensorApplySettingsRequestSignal[T]): The settings request that failed.
            reason (Optional[Exception]): The reason for the failure, if any.
        """
        RequestFailedSignal.__init__(self, source, reason=reason)
        self.request: SensorApplySettingsRequestSignal[T] = request


class CameraApplySettingsRequestFailedSignal(SensorApplySettingsRequestFailedSignal[Camera]):
    """Signal indicating that a request to apply camera settings has failed."""


#
# Request trigger
#


class SensorTriggerRequestSignal(NodeSignal, Generic[T]):
    """Represents a signal request to trigger a sensor.

    Attributes:
        sensor (Optional[T]): The specific sensor instance, if any, to trigger.

        timestamp (Optional[int]): Timestamp associated with the trigger request to identify
                                   produced images.

        adjust_acquisition_mode (bool): Specifies if the acquisition mode should be adjusted.
    """

    def __init__(
        self, source: Node, sensor: Optional[T], timestamp: Optional[int], adjust_acquisition_mode: bool = False
    ) -> None:
        """Initializes a SensorTriggerRequestSignal instance.

        Args:
            source (Node):
                The node that initiates the signal.

            sensor (Optional[T]):
                The sensor to be triggered. If None, all sensors are triggered simultaneously.

            timestamp (Optional[int]):
                Optional timestamp that can be used to identify the resulting item and link it
                to the trigger.

            adjust_acquisition_mode (bool, optional):
                If True, adjusts the acquisition mode. Defaults to False.
        """
        NodeSignal.__init__(self, source)
        self.sensor: Optional[T] = sensor
        self.timestamp: Optional[int] = timestamp
        self.adjust_acquisition_mode: bool = adjust_acquisition_mode


class CameraTriggerSignal(SensorTriggerRequestSignal[Camera]):
    """Represents a signal request to trigger a camera."""


class SensorTriggerRequestSignalFailed(RequestFailedSignal, Generic[T]):
    """Signal indicating that a sensor trigger request has failed."""

    def __init__(
        self, source: Node, request: SensorTriggerRequestSignal[T], reason: Optional[Exception] = None
    ) -> None:
        """Signal indicating that a sensor trigger request has failed.

        Args:
            source (Node): The node that initiated the signal.
            request (SensorTriggerRequestSignal[T]): The trigger request that failed.
            reason (Optional[Exception]): The reason for the failure, if any.
        """
        RequestFailedSignal.__init__(self, source, reason=reason)
        self.request: SensorTriggerRequestSignal[T] = request


class CameraTriggerSignalFailed(SensorTriggerRequestSignalFailed[Camera]):
    """Signal indicating that a camera trigger request has failed."""


#
# Request (re)connect
#


class SensorRequestReconnectSignal(NodeSignal, Generic[T]):
    """Signal requesting a sensor to reconnect."""

    def __init__(self, source: Node, sensor: T) -> None:
        """Signal requesting a sensor to reconnect.

        Args:
            source (Node): The node that initiated the signal.
            sensor (T): The sensor to be reconnected.
        """
        NodeSignal.__init__(self, source)
        self.sensor: T = sensor


class CameraRequestReconnectSignal(SensorRequestReconnectSignal[Camera]):
    """Signal requesting a camera to reconnect."""


class SensorRequestReconnectFailedSignal(RequestFailedSignal, Generic[T]):
    """Signal indicating that a sensor reconnect request has failed."""

    def __init__(
        self, source: Node, request: SensorRequestReconnectSignal[T], reason: Optional[Exception] = None
    ) -> None:
        """Signal indicating that a sensor reconnect request has failed.

        Args:
            source (Node): The node that initiated the signal.
            request (SensorRequestReconnectSignal[T]): The reconnect request that failed.
            reason (Optional[Exception]): The reason for the failure, if any.
        """
        RequestFailedSignal.__init__(self, source, reason=reason)
        self.request: SensorRequestReconnectSignal[T] = request


class CameraRequestReconnectFailedSignal(SensorRequestReconnectFailedSignal[Camera]):
    """Signal requesting a camera to reconnect."""


#
# Request disconnect
#


class SensorRequestDisconnectSignal(NodeSignal, Generic[T]):
    """Signal requesting a sensor to disconnect."""

    def __init__(self, source: Node, sensor: T) -> None:
        """Signal requesting a sensor to disconnect.

        Args:
            source (Node): The node that initiated the signal.
            sensor (T): The sensor to be disconnected.
        """
        NodeSignal.__init__(self, source)
        self.sensor: T = sensor


class CameraRequestDisconnectSignal(SensorRequestDisconnectSignal[Camera]):
    """Signal requesting a camera to disconnect."""


class SensorRequestDisconnectFailedSignal(RequestFailedSignal, Generic[T]):
    """Signal indicating that a sensor reconnect request has failed."""

    def __init__(
        self, source: Node, request: SensorRequestDisconnectSignal[T], reason: Optional[Exception] = None
    ) -> None:
        """Signal indicating that a sensor reconnect request has failed.

        Args:
            source (Node): The node that initiated the signal.
            request (SensorRequestReconnectSignal[T]): The reconnect request that failed.
            reason (Optional[Exception]): The reason for the failure, if any.
        """
        RequestFailedSignal.__init__(self, source, reason=reason)
        self.request: SensorRequestDisconnectSignal[T] = request


class CameraRequestDisconnectFailedSignal(SensorRequestDisconnectFailedSignal[Camera]):
    """Signal indicating that a camera disconnect request has failed."""
